"""lg-rez / blocs / Outils divers et variés

Récupération d'objets Discord, décorateurs pour commandes, structures
d'interaction dans les channels, utilitaires d'emojis, de date/temps,
de formatage...

"""

import asyncio
import datetime
import functools
import re
import warnings

import discord
import discord.utils
from discord.ext import commands
import unidecode

from lgrez import config, bdd
from lgrez.bdd import *
# on importe toutes les tables, plus simple pour y accéder depuis des
# réactions etc (via eval_accols)


# ---------------------------------------------------------------------------
# Utilitaires de récupération d'objets Discord (détectent les mentions)
# ---------------------------------------------------------------------------

#: Raccourci pour :func:`discord.utils.get`
get = discord.utils.get


def _find_by_mention_or_name(collec, val, pattern=None, must_be_found=False,
                             raiser=None):
    """Utilitaire pour la suite : trouve <val> dans <collec>

    [pattern]       Motif RegEx à utiliser pour la recherche
    [must_be_found] Si True, raise une ValueError si <val> est introuvable
    [raiser]        Nom de la fonction à envoyer dans l'exception si
                    introuvable
    """
    if not val:
        item = None
    elif pattern and (match := re.search(pattern, val)):
        item = get(collec, id=int(match.group(1)))
    else:
        item = get(collec, name=val)

    if must_be_found and not item:
        if raiser is None:
            raiser = "tools._find_by_mention_or_name"
        raise ValueError(f"{raiser} : Élément '{val}' introuvable")

    return item


def channel(nom, must_be_found=True):
    """Renvoie l'objet associé au salon ``#nom``.

    Args:
        nom (:class:`str`): nom du channel (texte/vocal/catégorie) ou sa
            mention (détection directe par regex)
        must_be_found (:class:`bool`): si ``True`` (défaut), raise une
            :exc:`ValueError` si le channel ``#nom`` n'existe pas
            (si ``False``, renvoie ``None``)

    Returns:
        :class:`discord.abc.GuildChannel`
    """
    return _find_by_mention_or_name(
        config.guild.channels, nom, pattern="<#([0-9]{18})>",
        must_be_found=must_be_found, raiser="tools.channel"
    )


def role(nom, must_be_found=True):
    """Renvoie l'objet associé au rôle ``@&nom``.

    Args:
        nom (:class:`str`): nom du rôle ou sa mention
            (détection directe par regex)
        must_be_found (:class:`bool`): si ``True`` (défaut), raise une
            :exc:`ValueError` si le channel ``@&nom`` n'existe pas
            (si ``False``, renvoie ``None``)

    Returns:
        :class:`discord.Role`
    """
    return _find_by_mention_or_name(
        config.guild.roles, nom, pattern="<@&([0-9]{18})>",
        must_be_found=must_be_found, raiser="tools.role"
    )


def member(nom, must_be_found=True):
    """Renvoie l'objet associé au membre ``@nom``.

    Args:
        nom (:class:`str`): nom du joueur ou sa mention
            (détection directe par regex)
        must_be_found (:class:`bool`): si ``True`` (défaut),
            raise une :exc:`ValueError` si le membre ``@nom`` n'existe pas
            (si ``False``, renvoie ``None``)

    Returns:
        :class:`discord.Member`
    """
    return _find_by_mention_or_name(
        config.guild.members, nom, pattern="<@!([0-9]{18})>",
        must_be_found=must_be_found, raiser="tools.member"
    )


def emoji(nom, must_be_found=True):
    """Renvoie l'objet associé à l'emoji ``:nom:``.

    Args:
        nom (:class:`str`): nom de l'emoji (texte/vocal/catégorie)
            ou son utilisation (détection directe par regex)
        must_be_found (:class:`bool`): si ``True`` (défaut), raise une
            :exc:`ValueError` si l'emoji ``:nom:`` n'existe pas
            (si ``False``, renvoie ``None``)

    Returns:
        :class:`discord.Emoji`
    """
    return _find_by_mention_or_name(
        config.guild.emojis, nom, pattern="<:.*:([0-9]{18})>",
        must_be_found=must_be_found, raiser="tools.emoji"
    )


# Appel aux MJs
def mention_MJ(arg):
    """Renvoie la mention ou le nom du rôle MJ

        - Si le joueur n'est pas un MJ, renvoie la mention de
          :attr:`config.Role.mj`
        - Sinon, renvoie son nom (pour ne pas rameuter tout le monde).

    Args:
        arg (:class:`~discord.Member` |\
            :class:`~discord.ext.commands.Context`):
            membre ou contexte d'un message envoyé par un membre

    Returns:
        :class:`str`
    """
    member = arg.author if isinstance(arg, commands.Context) else arg
    if (isinstance(member, discord.Member)
        and member.top_role >= config.Role.mj):
        # Pas un webhook et (au moins) MJ
        return f"@{config.Role.mj.name}"
    else:
        return config.Role.mj.mention


# ---------------------------------------------------------------------------
# Exceptions
# ---------------------------------------------------------------------------

class CommandExit(RuntimeError):
    """Fin de commande demandée.

    Lever cette exception force l'arrêt immédiat d'une commande,
    et empêche le bot de réagir à nouveau.

    Dérive de :exc:`RuntimeError`.
    """
    pass


# ---------------------------------------------------------------------------
# Décorateurs pour les différentes commandes, en fonction de leur usage
# ---------------------------------------------------------------------------

#: Décorateur pour commande (:func:`discord.ext.commands.check`) :
#: commande exécutable uniquement par un :attr:`MJ <.config.Role.mj>`
#: ou un webhook (tâche planifiée)
mjs_only = commands.check_any(
    commands.check(lambda ctx: ctx.message.webhook_id),
    commands.has_role(config.Role.get_raw("mj"))        # nom du rôle
)

#: Décorateur pour commandes d'IA (:func:`discord.ext.commands.check`) :
#: commande exécutable par un :attr:`MJ <.config.Role.mj>`, un
#: :attr:`Rédacteur <.config.Role.redacteur>` ou un webhook (tâche planifiée)
mjs_et_redacteurs = commands.check_any(
    mjs_only,
    commands.has_role(config.Role.get_raw("redacteur"))
)

#: Décorateur pour commande (:func:`discord.ext.commands.check`) :
#: commande exécutable uniquement par un joueur,
#: :attr:`vivant <.config.Role.joueur_en_vie>` ou
#: :attr:`mort <.config.Role.joueur_mort>`.
joueurs_only = commands.has_any_role(
    config.Role.get_raw("joueur_en_vie"),
    config.Role.get_raw("joueur_mort")
)

#: Décorateur pour commande (:func:`discord.ext.commands.check`) :
#: commande exécutable uniquement par un
#: :attr:`joueur vivant <.config.Role.joueur_en_vie>`
vivants_only = commands.has_role(config.Role.get_raw("joueur_en_vie"))


def private(callback):
    """Décorateur : commande utilisable dans son chan privé uniquement.

    Lors d'une invocation de la commande décorée hors d'un channel privé
    (commençant par :attr:`config.private_chan_prefix`), supprime le
    message d'invocation et exécute la commande dans le channel privé
    de l'invoqueur.

    Ce décorateur n'est utilisable que sur une commande définie dans un
    Cog. Si le joueur ayant utilisé la commande n'a pas de chan privé
    (pas en base), raise une :exc:`ValueError`.

    Utilisable en combinaison avec :func:`.joueurs_only` et
    :func:`.vivants_only` (pas avec les autres attention, vu que seuls
    les joueurs ont un channel privé).
    """
    @functools.wraps(callback)
    async def new_callback(self, ctx, *args, **kwargs):
        if not ctx.channel.name.startswith(config.private_chan_prefix):
            await ctx.message.delete()
            one_command.remove_from_in_command(ctx)
            # chan dans le contexte d'appel = chan privé
            ctx.channel = Joueur.from_member(ctx.author).private_chan
            await ctx.send(
                f"{quote(ctx.message.content)}\n"
                f"{ctx.author.mention} :warning: Cette commande est interdite"
                f" en dehors de ta conv privée ! :warning:\n"
                f"J'ai supprimé ton message, et j'exécute la commande ici :"
            )
            one_command.add_to_in_command(ctx)
        # Dans tous les cas, appelle callback (avec le contexte modifié)
        return await callback(self, ctx, *args, **kwargs)

    return new_callback


# ---------------------------------------------------------------------------
# Commandes d'interaction avec les joueurs : input, boucles, confirmation...
# ---------------------------------------------------------------------------

# Commande générale, à utiliser à la place de bot.wait_for('message', ...)
async def wait_for_message(check, trigger_on_commands=False, chan=None):
    """Attend le premier message reçu rencontrant les critères demandés.

    Surcouche de :meth:`discord.ext.commands.Bot.wait_for` permettant
    d'ignorer les commandes et de réagir au mot-clé ``stop``.

    Args:
        check (Callable[:class:`discord.Message` -> :class:`bool`]):
            fonction validant ou non chaque message.
        trigger_on_commands (bool): si ``False`` (défaut), un message
            respectant ``check`` sera ignoré si c'est une commande.
        chan (discord.TextChannel): le channel dans lequel le message
            est attendu (si applicable). Si ``None``, les messages
            d'arrêt (:obj:`config.stop_keywords`) peuvent ne pas être
            détectés (la fonction émettera un warning en ce sens).

    Returns:
        :class:`discord.Message`

    Raises:
        .CommandExit: si le message est un des :obj:`.config.stop_keywords`
            (insensible à la casse), même si il respecte ``check``
    """
    stop_keywords = [kw.lower() for kw in config.stop_keywords]

    if chan:
        def stop_check(m):
            return (m.channel == chan and m.content.lower() in stop_keywords)
    else:
        warnings.warn("lgrez.tools.wait_for_message called with `chan=None`, "
                      "stop messages may be ignored.", stacklevel=2)
        def stop_check(m):
            return False

    if trigger_on_commands:
        # on trigger en cas de STOP
        def trig_check(m):
            return (check(m) or stop_check(m))
    else:
        def trig_check(m):
            # on ne trigger pas sur les commandes et on trigger en cas de STOP
            return ((check(m)
                     and not m.content.startswith(config.bot.command_prefix))
                    or stop_check(m))

    message = await config.bot.wait_for('message', check=trig_check)
    if message.content.lower() in stop_keywords:
        if message.author == config.bot.user:
            raise CommandExit(ital("(Arrêt commande précédente)"))
        else:
            raise CommandExit("Arrêt demandé")
    else:
        return message


# Raccourci pratique
async def wait_for_message_here(ctx, trigger_on_commands=False):
    """Attend et renvoie le premier message reçu dans <ctx>.

    Surcouche de :func:`.wait_for_message` filtrant uniquement les
    messages envoyés dans ``ctx.channel`` par quelqu'un d'autre que
    le bot.

    Args:
        ctx (discord.ext.commands.Context): contexte d'une commande.
        trigger_on_commands: passé directement à
            :func:`.wait_for_message`.

    Returns:
        :class:`discord.Message`
    """
    def trig_check(message):
        return (message.channel == ctx.channel
                and message.author != ctx.bot.user)

    message = await wait_for_message(check=trig_check, chan=ctx.channel,
                                     trigger_on_commands=trigger_on_commands)
    return message


# Permet de boucler question -> réponse tant que la réponse ne
# vérifie pas les critères nécessaires
async def boucle_message(chan, in_message, condition_sortie, rep_message=None):
    """Boucle question/réponse jusqu'à qu'une condition soit vérifiée.

    Args:
        chan (discord.TextChannel): salon dans lequel lancer la boucle.
        condition_sortie (Callable[:class:`discord.Message` -> :class:`bool`]):
            fonction validant ou non chaque message.
        in_message (str): si défini, message à envoyer avant la boucle.
        rep_message (str): si défini, permet de définir un message de
            boucle différent de ``in_message`` (identique si ``None``).
            Doit être défini si ``in_message`` n'est pas défini.

    Returns:
        :class:`discord.Message`
    """
    if not rep_message:
        rep_message = in_message
    if not rep_message:
        raise ValueError("tools.boucle_message : `in_message` ou "
                         "`rep_message` doit être défini !")

    def check_chan(m):
        # Message envoyé pas par le bot et dans le bon chan
        return m.channel == chan and m.author != config.bot.user

    if in_message:
        await chan.send(in_message)
    rep = await wait_for_message(check_chan, chan=chan)
    while not condition_sortie(rep):
        await chan.send(rep_message)
        rep = await wait_for_message(check_chan, chan=chan)

    return rep


async def boucle_query(ctx, table, col=None, cible=None, filtre=None,
                       sensi=0.5, direct_detector=None, message=None):
    """Fait trouver à l'utilisateur une entrée de BDD d'après son nom.

    Args:
        ctx (discord.ext.commands.Context): contexte d'une commande.
        table (.bdd.base.TableMeta): table dans laquelle rechercher.
        col (~sqlalchemy.schema.Column): colonne dans laquelle rechercher
            (passé à :meth:`~.bdd.base.TableMeta.find_nearest`).
        cible (str): premier essai de cible (donnée par le joueur dans
            l'appel à une commande, par exemple).
        filtre: passé à :meth:`~.bdd.base.TableMeta.find_nearest`.
        sensi (float): sensibilité de la recherche (voir
            :meth:`~.bdd.TableMeta.find_nearest`).
        direct_detector (Callable[str] -> :attr:`table` | ``None``):
            pré-détecteur éventuel, appellé sur l'entrée utilisateur
            avant :meth:`~.bdd.TableMeta.find_nearest` ; si cette
            fonction renvoie un résultat, il est immédiatement renvoyé.
        message (str): si défini (et ``cible`` non défini), message à
            envoyer avant la boucle.

    Returns:
        Instance de :attr:`table` sélectionnée

    Attend que le joueur entre un nom, et boucle 5 fois au max
    (avant de l'insulter et de raise une erreur) pour chercher
    l'entrée la plus proche.
    """
    if message and not cible:
        await ctx.send(message)

    for i in range(5):
        if i == 0 and cible:
            # Au premier tour, si on a donné une cible
            rep = cible
        else:
            mess = await wait_for_message_here(ctx)
            rep = mess.content.strip("()[]{}<>")    # dézèlificateur

        # Détection directe
        if direct_detector:
            dir = direct_detector(rep)
            if dir:
                return dir

        # Sinon, recherche au plus proche
        nearest = table.find_nearest(rep, col=col, sensi=sensi, filtre=filtre,
                                     solo_si_parfait=False,
                                     match_first_word=True)

        if not nearest:
            await ctx.send("Aucune entrée trouvée, merci de réessayer :")

        elif len(nearest) == 1:         # Une seule correspondance
            result, score = nearest[0]
            if score == 1:          # parfait
                return result

            mess = await ctx.send("Je n'ai trouvé qu'une correspondance : "
                                  f"{bold(result)}.\nÇa part ?")
            if await yes_no(mess):
                return result
            else:
                await ctx.send("Bon d'accord, alors qui ?")

        else:
            text = ("Les résultats les plus proches de ton entrée "
                    "sont les suivants : \n")
            for i, (result, score) in enumerate(nearest[:10]):
                text += f"{emoji_chiffre(i + 1)}. {result} \n"
            mess = await ctx.send(
                text + ital("Tu peux les choisir en réagissant à ce "
                            "message, ou en répondant au clavier.")
            )
            n = await choice(mess, min(10, len(nearest)))
            return nearest[n - 1][0]

    await ctx.send("Et puis non, tiens !\nhttps://giphy.com/gifs/fuck-you-"
                   "middle-finger-ryan-stiles-x1kS7NRIcIigU")
    raise RuntimeError("Le joueur est trop con, je peux rien faire")


async def boucle_query_joueur(ctx, cible=None, message=None,
                              sensi=0.5, filtre=None):
    """Retourne un joueur (entrée de BDD) d'après son nom.

    Args:
        ctx (discord.ext.commands.Context): contexte d'une commande.
        cible (str): premier essai de cible (donnée par le joueur dans
            l'appel à une commande, par exemple).
        message (str): si défini (et ``cible`` non défini), message à
            envoyer avant la boucle.
        sensi (float): sensibilité de la recherche (voir
            :meth:`~.bdd.TableMeta.find_nearest`).
        filtre: passé à :meth:`~.bdd.TableMeta.find_nearest`.

    Returns:
        :class:`.bdd.Joueur`

    Attend que le joueur entre un nom de joueur, et boucle 5 fois au
    max (avant de l'insulter et de raise une erreur) pour chercher le
    plus proche joueur dans la table :class:`.bdd.Joueur`.
    """
    # Détection directe par ID / nom exact
    def direct_detector(rep):
        mem = member(rep, must_be_found=False)
        if mem:
            try:                    # Récupération du joueur
                return Joueur.from_member(mem)
            except ValueError:          # pas inscrit en base
                pass

        return None

    res = await boucle_query(ctx, Joueur, col=Joueur.nom, cible=cible,
                             sensi=sensi, filtre=filtre,
                             direct_detector=direct_detector,
                             message=message)
    return res


# Récupère un input par réaction
async def wait_for_react_clic(message, emojis={}, *, process_text=False,
                              text_filter=None, first_text=None,
                              post_converter=None,
                              trigger_all_reacts=False,
                              trigger_on_commands=False):
    """Ajoute des reacts à un message et attend une interaction.

    Args:
        message (discord.Message): message où ajouter les réactions.
        emojis (:class:`list` | :class:`dict`): reacts à ajouter,
            éventuellement associés à une valeur qui sera retournée
            si clic sur l'emoji.
        process_text (bool): si ``True``, détecte aussi la réponse par
            message et retourne le texte du message (défaut : ``False``).
        text_filter (Callable[:class:`str` -> :class:`bool`]): si
            ``process_text``, ne réagit qu'aux messages pour lesquels
            ``text_filter(message)`` renvoie ``True`` (défaut : tous).
        first_text (str): si ``process_text``, texte considéré comme la
            première réponse textuelle reçue (si il vérifie
            ``text_filter``, les emojis ne sont pas ajoutés et cette
            fonction retourne directement).
        post_converter (Callable[:class:`str` -> Any]): si
            ``process_text`` et que l'argument est défini, le message
            détecté est passé dans cette fonction avant d'être renvoyé.
        trigger_all_reacts (bool): si ``True``, détecte l'ajout de
            toutes les réactions (pas seulement celles dans ``emojis``)
            et renvoie l'emoji directement si il n'est pas dans
            ``emojis`` (défaut : ``False``).
        trigger_on_commands (bool): passé à :func:`.wait_for_message`.

    Returns:
        - :class:`str` -- représentant
            - le nom de l'emoji si ``emojis`` est une liste et clic sur
              une des reacts, ou si ``trigger_all_reacts`` vaut ``True``
              et ajout d'une autre react ;
            - le message reçu si ``process_text`` vaut ``True``, que
              ``post_converter`` n'est pas défini et réaction à un
              message ;
        - Any -- représentant
            - la valeur associée si ``emojis`` est un dictionnaire et
              clic sur une des reacts ;
            - la valeur retournée par ``post_converter`` si il est
              défini, que ``process_text`` vaut ``True`` et réaction
              à un message.
    """
    if not isinstance(emojis, dict):
        # Si emoji est une liste, on en fait un dictionnaire
        emojis = {emoji: emoji for emoji in emojis}

    if text_filter is None:
        def text_filter(text):
            return True

    if process_text and first_text:
        if text_filter(first_text):     # passe le filtre
            return post_converter(first_text) if post_converter else first_text

    try:
        # Si une erreur dans ce bloc, on supprime les emojis
        # du message (sinon c'est moche)
        for emoji in emojis:
            try:
                await message.add_reaction(emoji)
            except discord.errors.HTTPException:
                await message.channel.send(f"*Emoji {emoji} inconnu, ignoré*")

        emojis_names = {emoji.name if hasattr(emoji, "name")
                        else emoji: emoji for emoji in emojis}

        def react_check(react):
            # Check REACT : bon message, bon emoji, et pas react du bot
            name = react.emoji.name
            return (react.message_id == message.id
                    and react.user_id != config.bot.user.id
                    and (trigger_all_reacts or name in emojis_names))

        react_task = asyncio.create_task(
            config.bot.wait_for('raw_reaction_add', check=react_check)
        )

        if process_text:
            # Check MESSAGE : bon channel, pas du bot, et filtre
            def message_check(mess):
                return (mess.channel == message.channel
                        and mess.author != config.bot.user
                        and text_filter(mess.content))
        else:
            # On process DANS TOUS LES CAS, mais juste pour détecter
            # les stop_keywords si process_text == False
            def message_check(mess):
                return False

        mess_task = asyncio.create_task(
            wait_for_message(check=message_check, chan=message.channel,
                             trigger_on_commands=trigger_on_commands)
        )

        done, pending = await asyncio.wait([react_task, mess_task],
                                           return_when=asyncio.FIRST_COMPLETED)
        # Le bot attend ici qu'une des deux tâches aboutisse
        for task in pending:
            task.cancel()
        done_task = next(iter(done))        # done = tâche aboutie

        if done_task == react_task:         # Réaction
            emoji = done_task.result().emoji
            if trigger_all_reacts and emoji.name not in emojis_names:
                ret = emoji
            else:
                ret = (emojis.get(emoji)
                       or emojis.get(emojis_names.get(emoji.name)))

            for emoji in emojis:
                # On finit par supprimer les emojis mis par le bot
                await message.remove_reaction(emoji, config.bot.user)

        else:       # Réponse par message / STOP
            mess = done_task.result().content
            ret = post_converter(mess) if post_converter else mess
            await message.clear_reactions()

    except Exception:
        await message.clear_reactions()
        raise

    return ret


async def yes_no(message, first_text=None):
    """Demande une confirmation / question fermée à l'utilisateur.

    Surcouche de :func:`wait_for_react_clic` : ajoute les reacts
    ✅ et ❎ à un message et renvoie ``True`` ou ``False`` en fonction
    de l'emoji cliqué OU de la réponse textuelle détectée.

    Args:
        message (discord.Message): message où ajouter les réactions.
        first_text (str): passé à :func:`wait_for_react_clic`.

    Réponses textuelles reconnues :
        - Pour ``True`` : ``["oui", "o", "yes", "y", "1", "true"]``
        - Pour ``False`` : ``["non", "n", "no", "n", "0", "false"]``

    ainsi que toutes leurs variations de casse.

    Returns:
        :class:`bool`
    """
    yes_words = ["oui", "o", "yes", "y", "1", "true"]
    yes_no_words = yes_words + ["non", "n", "no", "n", "0", "false"]
    return await wait_for_react_clic(
        message, emojis={"✅": True, "❎": False}, process_text=True,
        first_text=first_text,
        text_filter=lambda s: s.lower() in yes_no_words,
        post_converter=lambda s: s.lower() in yes_words,
    )

yes_no_maybe_i_dont_know_can_you_repeat_the_question = yes_no


async def choice(message, N, start=1, *, additionnal={}):
    """Demande à l'utilisateur de choisir entre plusieurs options numérotées.

    Surcouche de :func:`wait_for_react_clic` : ajoute des reacts
    chiffres (1️⃣, 2️⃣, 3️⃣...) et renvoie le numéro cliqué OU détecté
    par réponse textuelle.

    Args:
        message (discord.Message): message où ajouter les réactions.
        N (int): chiffre jusqu'auquel aller, inclus (``<= 10``).
        start (int): chiffre auquel commencer (entre ``0`` et ``N``,
            défaut ``1``).
        additionnal (dict[:class:`discord.Emoji` | :class:`str`, Any]):
            emojis optionnels à ajouter après les chiffres et valeur
            renvoyée si cliqué.

    Réponses textuelles reconnues : chiffres entre ``start`` et ``N``.

    Returns:
        :class:`int` (ou la valeur associée si emoji choisi dans
        ``additionnal``)
    """
    emojis = {emoji_chiffre(i): i for i in range(start, N + 1)}
    emojis.update(additionnal)
    return await wait_for_react_clic(
        message, emojis=emojis, process_text=True,
        text_filter=lambda s: s.isdigit() and start <= int(s) <= N,
        post_converter=int,
    )


async def sleep(chan, tps):
    """Attend un temps donné en avertissant l'utilisateur.

    Pause l'exécution d'une commande en affichant l'indicateur *typing*
    ("*LGBot est en train d'écrire...*") sur un salon.

    Permat d'afficher plusieurs messages d'affillée en laissant le temps
    de lire, tout en indiquant que le bot n'a pas fini d'écrire.

    Args:
        chan (discord.abc.Messageable): salon / contexte /... sur lequel
            attendre.
        tps (float): temps à attendre, en secondes.
    """
    async with chan.typing():
        await asyncio.sleep(tps)


# ---------------------------------------------------------------------------
# Utilitaires d'emojis
# ---------------------------------------------------------------------------

def montre(heure=None):
    """Renvoie l'emoji horloge le plus proche d'une heure donnée.

    Args:
        heure (str): heure à représenter, au format ``"XXh"`` ou
            ``"XXhMM"`` (défaut : heure actuelle).

    Returns:
        :class:`str` (🕧, 🕓, 🕝...)
    """
    if heure and isinstance(heure, str):
        heure, minute = heure.split("h")
        heure = int(heure) % 12
        minute = int(minute) % 60 if minute else 0
    else:
        now = datetime.datetime.now()
        heure = now.hour % 12
        minute = now.minute

    if minute >= 45:
        heure = (heure + 1) % 12

    if 15 < minute < 45:        # Demi heure
        L = ["\N{CLOCK FACE TWELVE-THIRTY}",    "\N{CLOCK FACE ONE-THIRTY}",
             "\N{CLOCK FACE TWO-THIRTY}",       "\N{CLOCK FACE THREE-THIRTY}",
             "\N{CLOCK FACE FOUR-THIRTY}",      "\N{CLOCK FACE FIVE-THIRTY}",
             "\N{CLOCK FACE SIX-THIRTY}",       "\N{CLOCK FACE SEVEN-THIRTY}",
             "\N{CLOCK FACE EIGHT-THIRTY}",     "\N{CLOCK FACE NINE-THIRTY}",
             "\N{CLOCK FACE TEN-THIRTY}",       "\N{CLOCK FACE ELEVEN-THIRTY}"]
    else:                       # Heure pile
        L = ["\N{CLOCK FACE TWELVE OCLOCK}",    "\N{CLOCK FACE ONE OCLOCK}",
             "\N{CLOCK FACE TWO OCLOCK}",       "\N{CLOCK FACE THREE OCLOCK}",
             "\N{CLOCK FACE FOUR OCLOCK}",      "\N{CLOCK FACE FIVE OCLOCK}",
             "\N{CLOCK FACE SIX OCLOCK}",       "\N{CLOCK FACE SEVEN OCLOCK}",
             "\N{CLOCK FACE EIGHT OCLOCK}",     "\N{CLOCK FACE NINE OCLOCK}",
             "\N{CLOCK FACE TEN OCLOCK}",       "\N{CLOCK FACE ELEVEN OCLOCK}"]
    return L[heure]


def emoji_chiffre(chiffre, multi=False):
    """Renvoie l'emoji / les emojis chiffre correspondant à un chiffre/nombre.

    Args:
        chiffre (int): chiffre/nombre à représenter.
        multi (bool): si ``True``, ``chiffre`` peut être n'importe quel
            entier positif, dont les chiffres seront convertis
            séparément ; sinon (par défaut), ``chiffre`` doit être un
            entier entre ``0`` et ``10``.

    Returns:
        :class:`str` (0️⃣, 1️⃣, 2️⃣...)
    """
    if isinstance(chiffre, int) and 0 <= chiffre <= 10:
        return ["0️⃣", "1️⃣", "2️⃣", "3️⃣", "4️⃣",
                "5️⃣", "6️⃣", "7️⃣", "8️⃣", "9️⃣", "🔟"][chiffre]
    elif multi and str(chiffre).isdigit():
        return "".join([emoji_chiffre(int(chr)) for chr in str(chiffre)])
    else:
        raise ValueError("L'argument de tools.emoji_chiffre doit être un "
                         "entier entre 0 et 10 OU un entier positif avec "
                         "multi=True")


def super_chiffre(chiffre, multi=False):
    """Renvoie le(s) caractère(s) exposant correspondant à un chiffre/nombre.

    Args:
        chiffre (int): chiffre/nombre à représenter.
        multi (bool): si ``True``, ``chiffre`` peut être n'importe quel
            entier positif, dont les chiffres seront convertis
            séparément ; sinon (par défaut), ``chiffre`` doit être un
            entier entre ``0`` et ``9``.

    Returns:
        :class:`str` (⁰, ¹, ²...)
    """
    if isinstance(chiffre, int) and 0 <= chiffre <= 9:
        return ["⁰", "¹", "²", "³", "⁴", "⁵", "⁶", "⁷", "⁸", "⁹"][chiffre]
    elif multi and str(chiffre).isdigit():
        return ''.join([super_chiffre(int(chr)) for chr in str(chiffre)])
    else:
        raise ValueError("L'argument de tools.super_chiffre doit être un "
                         "entier entre 0 et 9 OU un entier positif avec "
                         "multi=True")


def sub_chiffre(chiffre, multi=False):
    """Renvoie le(s) caractère(s) indice correspondant à un chiffre/nombre.

    Args:
        chiffre (int): chiffre/nombre à représenter.
        multi (bool): si ``True``, ``chiffre`` peut être n'importe quel
            entier positif, dont les chiffres seront convertis
            séparément ; sinon (par défaut), ``chiffre`` doit être un
            entier entre ``0`` et ``9``.

    Returns:
        :class:`str` (₀, ₁, ₂...)
    """
    if isinstance(chiffre, int) and 0 <= chiffre <= 9:
        return ["₀", "₁", "₂", "₃", "₄", "₅", "₆", "₇", "₈", "₉"][chiffre]
    elif multi and str(chiffre).isdigit():
        return ''.join([sub_chiffre(int(c)) for c in str(chiffre)])
    else:
        raise ValueError("L'argument de tools.sub_chiffre doit être un "
                         "entier entre 0 et 9 OU un entier positif avec "
                         "multi=True")


# ---------------------------------------------------------------------------
# Utilitaires de date / temps, notemment liées aux horaires de jeu
# ---------------------------------------------------------------------------

def heure_to_time(heure):
    """Convertit l'écriture d'une heure en objet :class:`datetime.time`.

    Args:
        heure (str): heure au format ``HHh``, ``HHhMM`` ou ``HH:MM``.

    Returns:
        :class:`datetime.time`

    Raises:
        ValueError: conversion impossible (mauvais format)
    """
    try:
        if "h" in heure:
            hh, mm = heure.split("h")
        else:
            hh, mm = heure.split(":")
        return datetime.time(int(hh), int(mm) if mm else 0)
    except ValueError as exc:
        raise ValueError(f"Valeur \"{heure}\" non convertible "
                         "en temps") from exc


def time_to_heure(tps, sep="h", force_minutes=False):
    """Convertit un objet :class:`datetime.time` en heure.

    (version maison de :meth:`datetime.time.strftime`)

    Args:
        tps (datetime.time): temps à convertir.
        sep (str): séparateur heures / minutes à utiliser
            (défaut ``"h"``).
        force_minutes (bool): si ``False`` (défaut), les minutes
            ne sont indiquées que si différentes de ``0``.

    Returns:
        :class:`str` (``""`` si ``tps`` est ``None``)
    """
    if tps:
        if force_minutes or tps.minute > 0:
            return f"{tps.hour}{sep}{tps.minute:02}"
        else:
            return f"{tps.hour}{sep}"
    else:
        return ""


def next_occurence(tps):
    """Renvoie la prochaine occurence temporelle d'une heure donnée.

    Renvoie le prochain timestamp arrivant DANS LES HORAIRES DU JEU :
    entre :func:`.tools.fin_pause` et :func:`.tools.debut_pause`.

    Args:
        tps (datetime.time): heure dont on veut connaître
            la prochaine occurence.

    Returns:
        :class:`datetime.datetime`
    """
    now = datetime.datetime.now()
    jour = now.date()
    if tps < now.time():
        # Si plus tôt dans la journée que l'heure actuelle,
        # on réfléchit comme si on était demain
        jour += datetime.timedelta(days=1)

    test_dt = datetime.datetime.combine(jour, tps)
    if test_dt < debut_pause() and not en_pause():
        # Prochaine occurence avant la pause : OK
        return test_dt

    # Sinon, programmer après la pause
    finp = fin_pause()
    jour = finp.date()
    if tps < finp.time():
        # Si plus tôt dans la journée que l'heure de reprise,
        # on réfléchit comme si on était le lendemain
        jour += datetime.timedelta(days=1)

    return datetime.datetime.combine(jour, tps)


def debut_pause():
    """Renvoie le timestamp correspondant au prochain vendredi 19h.

    Returns:
        :class:`datetime.datetime`
    """
    pause_time = datetime.time(hour=19)
    pause_wday = 4          # Vendredi

    now = datetime.datetime.now()
    jour = now.date()
    if pause_time < now.time():
        # Si plus tôt dans la journée que l'heure actuelle,
        # on réfléchit comme si on était demain
        jour += datetime.timedelta(days=1)

    ddays = (pause_wday - jour.weekday()) % 7
    # Jour décalé du nombre de jours avant vendredi
    pause_jour = jour + datetime.timedelta(days=ddays)
    return datetime.datetime.combine(pause_jour, pause_time)


def fin_pause():
    """Renvoie le timestamp correspondant au prochain dimanche 19h.

    Returns:
        :class:`datetime.datetime`
    """
    reprise_time = datetime.time(hour=19)
    reprise_wday = 6        # Dimanche

    now = datetime.datetime.now()
    jour = now.date()
    if reprise_time < now.time():
        # Si plus tôt dans la journée que l'heure actuelle,
        # on réfléchit comme si on était demain
        jour += datetime.timedelta(days=1)

    ddays = (reprise_wday - jour.weekday()) % 7
    # Jour décalé du nombre de jours avant vendredi
    reprise_jour = jour + datetime.timedelta(days=ddays)
    return datetime.datetime.combine(reprise_jour, reprise_time)


def en_pause():
    """Détermine si le jeu est actuellement en pause hebdomadaire.

    Si il n'y a pas de pause (:func:`.fin_pause` = :func:`.debut_pause`),
    renvoie toujours ``False``.

    Returns:
        :class:`bool`
    """
    return fin_pause() < debut_pause()


# ---------------------------------------------------------------------------
# Split et log
# ---------------------------------------------------------------------------

def smooth_split(mess, N=1990, sep='\n', rep=''):
    """Sépare un message en une blocs moins longs qu'une limite donnée.

    Très utile pour envoyer des messages de (potentiellement) plus de
    2000 caractères (limitation Discord).

    Args:
        mess (str): message à découper.
        N (int): taille maximale des messages formés (défaut ``1990``,
            pour avoir un peu de marge par rapport à la limitation, et
            permettre d'entourer de ``````` par exemple)
        sep (str): caractères où séparer préférentiellement le texte
            (défaut : sauts de ligne). Si ``mess`` contient une
            sous-chaîne plus longue que ``N`` ne contenant pas ``sep``,
            le message sera tronqué à la limite.
        rep (str) : chaîne ajoutée à la fin de chaque message formé
            (tronqué du séparateur final) (défaut : aucune).

    Returns:
        :class:`list`\[:class:`str`\]
    """
    mess = str(mess)
    LM = []             # Liste des messages
    psl = 0             # indice du Précédent Saut de Ligne
    L = len(mess)
    while psl + N < L:
        if mess.count(sep, psl, psl + N + len(sep)):
            # +len(sep) parce que si sep est à la fin, on le dégage
            i = psl + N - mess[psl: psl + N + len(sep)][::-1].find(sep)
            # un peu sombre mais vrai, tkt frère
            LM.append(mess[psl: i] + rep)
            psl = i + 1     # on élimine le \n
        else:
            LM.append(mess[psl: psl + N] + rep)
            psl += N

    if psl < L:
        LM.append(mess[psl:])   # ce qui reste
    return LM


async def send_blocs(messageable, mess, *, N=1990, sep='\n', rep=''):
    """Envoie un message en le coupant en blocs si nécaissaire.

    Surcouche de :func:`.smooth_split` envoyant directement
    les messages formés.

    Args:
        messageable (discord.abc.Messageable): objet où envoyer le
            message (:class:`~discord.ext.commands.Context` ou
            :class:`~discord.TextChannel`).
        mess (str): message à envoyer
        N, sep, rep: passé à :func:`.smooth_split`.

    Returns:
        list[discord.Message]: La liste des messages envoyés.
    """
    messages = []
    for bloc in smooth_split(mess, N=N, sep=sep, rep=rep):
        messages.append(await messageable.send(bloc))

    return messages


async def send_code_blocs(messageable, mess, *, N=1990, sep='\n', rep='',
                          prefixe="", langage=""):
    """Envoie un (potentiellement long) message sous forme de bloc(s) de code.

    Équivalent de :func:`.send_blocs` avec formatage de chaque bloc
    dans un bloc de code.

    Args:
        messageable, mess, N, sep, rep: voir :func:`.send_blocs`.
        prefixe (str): texte optionnel à mettre hors des code blocs,
            au début du premier message.
        language (str): voir :func:`.code_bloc`.

    Returns:
        list[discord.Message]: La liste des messages envoyés.
    """
    mess = str(mess)

    if prefixe:
        prefixe = prefixe.rstrip() + "\n"

    blocs = smooth_split(prefixe + mess, N=N, sep=sep, rep=rep)

    messages = []
    for i, bloc in enumerate(blocs):
        if prefixe and i == 0:
            bloc = bloc[len(prefixe):]
            message = await messageable.send(
                prefixe + code_bloc(bloc, langage=langage))
        else:
            message = await messageable.send(code_bloc(bloc, langage=langage))
        messages.append(message)

    return messages


async def log(message, *, code=False, N=1990, sep='\n', rep='',
              prefixe="", langage=""):
    """Envoie un message dans le channel :attr:`config.Channel.logs`.

    Surcouche de :func:`.send_blocs` / :func:`.send_code_blocs`.

    Args:
        message (str): message à log.
        code (bool): si ``True``, log sous forme de bloc(s) de code
            (défaut ``False``).
        N, sep, rep: passé à :func:`.send_blocs` /
            :func:`.send_code_blocs`.
        prefixe: voir :func:`.send_code_blocs`, simplement ajouté avant
            ``message`` si ``code`` vaut ``False``.
        language: *identique à* :func:`.send_code_blocs`, ignoré
            si `code` vaut ``False``.

    Returns:
        list[discord.Message]: La liste des messages envoyés.
    """
    logchan = config.Channel.logs
    if code:
        return (await send_code_blocs(logchan, message, N=N, sep=sep, rep=rep,
                                      prefixe=prefixe, langage=langage))
    else:
        if prefixe:
            message = prefixe.rstrip() + "\n" + message
        return (await send_blocs(logchan, message, N=N, sep=sep, rep=rep))


# ---------------------------------------------------------------------------
# Autres fonctions diverses
# ---------------------------------------------------------------------------

async def create_context(member, content):
    """Génère le contexte associé au message d'un membre dans son chan privé.

    Args:
        member (discord.Member): membre dont on veut simuler l'action.
            **Doit être inscrit en base** (pour avoir un chan privé).
        content (str): message à "faire envoyer" au joueur,
            généralement une commande.

    Utile notemment pour simuler des commandes à partir de clics sur
    des réactions.

    Returns:
        :class:`discord.ext.commands.Context`
    """
    chan = Joueur.from_member(member).private_chan
    message = (await chan.history(limit=1).flatten())[0]
    # On a besoin de récupérer un message, ici le dernier de la conv privée
    message.author = member
    message.content = content
    ctx = await config.bot.get_context(message)
    return ctx


def remove_accents(text):
    """Enlève les accents d'un chaîne, mais conserve les caractères spéciaux.

    Version plus douce de ``unidecode.unidecode``, conservant notemment
    les emojis, ...

    Args:
        text (str): chaîne à désaccentuer.

    Returns:
        :class:`str`
    """
    p = re.compile("([À-ʲΆ-ת])")
    # Abracadabrax, c'est moche mais ça marche (source : tkt frère)
    return p.sub(lambda c: unidecode.unidecode(c.group()), text)


# Évaluation d'accolades
def eval_accols(rep, globals_=None, locals_=None, debug=False):
    """Replace chaque bloc entouré par des ``{}`` par leur évaluation Python.

    Args:
        globals_ (dict): variables globales du contexte d'évaluation
            (passé à :func:`eval`).
        locals_ (dict): variables locales du contexte d'évaluation
            (passé à :func:`eval`).
        debug (bool): si ``True``, insère le message d'erreur (type et
            texte de l'exception) dans le message à l'endroit où une
            exception est levée durant l'évaluation (défaut ``False``).

    Penser à passer les :func:`globals` et :func:`locals` si besoin.
    Généralement, il faut passer :func:`locals` qui contient ``ctx``,
    etc... mais pas :func:`globals` si on veut bénéficier de tous les
    modules importés dans ``tools.py``.
    """
    if globals_ is None:
        globals_ = globals()
    if locals_ is None:
        locals_ = globals_

    if "{" not in rep:          # Si pas d'expressions, on renvoie direct
        return rep

    evrep = ""                  # Réponse évaluée
    expr = ""                   # Expression à évaluer
    noc = 0                     # Nombre de { non appariés
    for car in rep:
        if car == "{":
            if noc:             # Expression en cours :
                expr += car         # on garde le {
            noc += 1
        elif car == "}":
            noc -= 1
            if noc:             # idem
                expr += car
            else:               # Fin d'une expression
                try:                # On essaie d'évaluer la chaîne
                    evrep += str(eval(expr, globals_, locals_))
                except Exception as e:
                    # Si erreur, on laisse {expr} non évaluée
                    evrep += "{" + expr + "}"
                    if debug:
                        evrep += code(f"->!!! {e} !!!")
                expr = ""
        elif noc:               # Expression en cours
            expr += car
        else:                   # Pas d'expression en cours
            evrep += car
    if noc:     # Si expression jamais finie (nombre impair de {)
        evrep += "{" + expr
    return evrep


# ---------------------------------------------------------------------------
# Utilitaires de formatage de texte
# ---------------------------------------------------------------------------

def bold(text):
    """Formate une chaîne comme texte en **gras** dans Discord.

    Args:
        text (str): chaîne à formater.

    Returns:
        :class:`str`
    """
    return f"**{text}**"


def ital(text):
    """Formate une chaîne comme texte en *italique* dans Discord.

    Args:
        text (str): chaîne à formater.

    Returns:
        :class:`str`
    """
    return f"*{text}*"


def soul(text):
    """Formate une chaîne comme texte souligné dans Discord.

    Args:
        text (str): chaîne à formater.

    Returns:
        :class:`str`
    """
    return f"__{text}__"


def strike(text):
    """Formate une chaîne comme texte barré dans Discord.

    Args:
        text (str): chaîne à formater.

    Returns:
        :class:`str`
    """
    return f"~~{text}~~"


def code(text):
    """Formate une chaîne comme ``code`` (inline) dans Discord.

    Args:
        text (str): chaîne à formater.

    Returns:
        :class:`str`
    """
    return f"`{text}`"


def code_bloc(text, langage=""):
    """Formate une chaîne comme un bloc de code dans Discord.

    Args:
        text (str): chaîne à formater.
        langage (str): langage du code, pour coloration syntaxique.

    Langages supportés (non exhaustif ?) : ``asciidoc``, ``autohotkey``,
    ``bash``, ``coffeescript``, ``cpp`` (C++), ``cs`` (C#), ``css``,
    ``diff``, ``fix``, ``glsl``, ``ini``, ``json``, ``md``, (markdown),
    ``ml``, ``prolog``, ``py``, ``tex``, ``xl``, ``xml``

    Returns:
        :class:`str`
    """
    return f"```{langage}\n{text}```"


def quote(text):
    """Formate une chaîne comme citation (inline) dans Discord.

    Args:
        text (str): chaîne à formater.

    Returns:
        :class:`str`
    """
    return f"> {text}"


def quote_bloc(text):
    """Formate une chaîne comme bloc de citation (multiline) dans Discord.

    Args:
        text (str): chaîne à formater.

    Returns:
        :class:`str`
    """
    return f">>> {text}"


def spoiler(text):
    """Formate une chaîne comme spoiler (cliquer pour afficher) dans Discord.

    Args:
        text (str): chaîne à formater.

    Returns:
        :class:`str`
    """
    return f"||{text}||"



import unittest
from unittest import mock

import sqlalchemy

from lgrez import config
from lgrez.bdd import base
from test import mock_discord, mock_bdd, mock_env



class TestBaseFunctions(unittest.IsolatedAsyncioTestCase):
    """Unit tests for lgrez.bdd.base functions."""

    def setUp(self):
        mock_discord.mock_config()

    def tearDown(self):
        mock_discord.unmock_config()

    def test__remove_accents(self):
        """Unit tests for base._remove_accents function."""
        # def _remove_accents(text)
        _remove_accents = base._remove_accents
        samples = {
            "": "",
            "allo": "allo",
            "AllO": "AllO",
            "énorme ÉNORME": "enorme ENORME",
            "àéïìùÀÉÏÌÙ‰": "aeiiuAEIIU‰",
            "énorme?♥♦♣♠": "enorme?♥♦♣♠",
        }
        for sample, result in samples.items():
            self.assertEqual(_remove_accents(sample), result)


    # Objets d'autodoc partiellement couverts seulement :
    # vérification que l'objet SQLAlchemy correspondant est correctement
    # appelé (à part doc et comment), pas que la doc est cohérente

    @mock.patch("sqlalchemy.Column")
    def test_autodoc_Column(self, col_patch):
        """Unit tests for base.autodoc_Column function."""
        # def autodoc_Column(*args, doc="", comment=None, **kwargs)
        autodoc_Column = base.autodoc_Column
        args = [mock.Mock() for i in range(5)]
        kwargs = {str(i): mock.Mock() for i in range(5)}
        col_patch.return_value.configure_mock(**{
            "type.python_type.__name__": "a"})
        col = autodoc_Column(*args, **kwargs)
        col_patch.assert_called_once_with(*args, **kwargs, comment=mock.ANY)
        self.assertIs(col, col_patch.return_value)

    @mock.patch("sqlalchemy.orm.relationship")
    def test_autodoc_ManyToOne(self, rsp_patch):
        """Unit tests for base.autodoc_ManyToOne function."""
        # def autodoc_ManyToOne(tablename, *args, doc="", **kwargs)
        autodoc_ManyToOne = base.autodoc_ManyToOne
        args = [mock.Mock() for i in range(5)]
        kwargs = {str(i): mock.Mock() for i in range(5)}
        rsp = autodoc_ManyToOne("ozy", *args, **kwargs)
        rsp_patch.assert_called_once_with("ozy", *args, **kwargs, doc=mock.ANY)
        self.assertIs(rsp, rsp_patch.return_value)

    @mock.patch("sqlalchemy.orm.relationship")
    def test_autodoc_OneToMany(self, rsp_patch):
        """Unit tests for base.autodoc_OneToMany function."""
        # def autodoc_OneToMany(tablename, *args, doc="", **kwargs)
        autodoc_OneToMany = base.autodoc_OneToMany
        args = [mock.Mock() for i in range(5)]
        kwargs = {str(i): mock.Mock() for i in range(5)}
        rsp = autodoc_OneToMany("ozy", *args, **kwargs)
        rsp_patch.assert_called_once_with("ozy", *args, **kwargs, doc=mock.ANY)
        self.assertIs(rsp, rsp_patch.return_value)

    @mock.patch("sqlalchemy.orm.relationship")
    def test_autodoc_ManyToMany(self, rsp_patch):
        """Unit tests for base.autodoc_ManyToMany function."""
        # def autodoc_ManyToMany(tablename, *args, doc="", **kwargs)
        autodoc_ManyToMany = base.autodoc_ManyToMany
        args = [mock.Mock() for i in range(5)]
        kwargs = {str(i): mock.Mock() for i in range(5)}
        rsp = autodoc_ManyToMany("ozy", *args, **kwargs)
        rsp_patch.assert_called_once_with("ozy", *args, **kwargs, doc=mock.ANY)
        self.assertIs(rsp, rsp_patch.return_value)


    @mock_env.patch_env(LGREZ_DATABASE_URI="oh!<3")
    @mock.patch("sqlalchemy.create_engine")
    @mock.patch("sqlalchemy.orm.sessionmaker")
    @mock.patch("lgrez.bdd.base.TableBase")
    def test_connect(self, tb_patch, sm_patch, ce_patch):
        """Unit tests for base.connect function."""
        # def connect()
        connect = base.connect
        del config.session, config.engine
        connect()
        ce_patch.assert_called_once_with("oh!<3", pool_pre_ping=True)
        self.assertEqual(config.engine, ce_patch.return_value)
        tb_patch.metadata.create_all.assert_called_once_with(config.engine)
        sm_patch.assert_called_once_with(bind=config.engine)
        self.assertEqual(config.session, sm_patch.return_value.return_value)



class TestTableMeta(unittest.IsolatedAsyncioTestCase):
    """Unit tests for lgrez.bdd.base.TableMeta methods."""

    def setUp(self):
        mock_discord.mock_config()

    def tearDown(self):
        mock_discord.unmock_config()

    def test___init__(self):
        """Unit tests for TableMeta.__init__ method."""
        # def __init__(cls, name, bases, dic, comment=None, **kwargs)
        __init__ = base.TableMeta.__init__
        cls = mock.MagicMock(base.TableMeta)
        cls.__doc__ = "hophop"

        # basic
        with mock.patch("sqlalchemy.ext.declarative.api"
                        ".DeclarativeMeta.__init__") as dm_patch:
            __init__(cls, "TestTab", (), {})
        self.assertEqual(cls.__tablename__, "testtabs")
        dm_patch.assert_called_once_with("TestTab", (), {}, comment="hophop")
        self.assertEqual(cls._attrs, {})
        self.assertTrue(cls.__doc__.startswith("hophop\n"))
        self.assertIn("testtabs", cls.__doc__)

        # complex
        cls.__doc__ = "haphap"
        bases = mock.Mock()
        dic = {
            "ah": 1,
            "allo": mock.Mock(),
            "coucou": mock.Mock(sqlalchemy.Column),
            "kuku": mock.Mock(sqlalchemy.orm.relationships.RelationshipProperty),
        }
        with mock.patch("sqlalchemy.ext.declarative.api"
                        ".DeclarativeMeta.__init__") as dm_patch:
            __init__(cls, "PoPoPo", bases, dic, "keuwa", a=2, z="a")
        self.assertEqual(cls.__tablename__, "popopos")
        dm_patch.assert_called_once_with("PoPoPo", bases, dic,
                                         comment="keuwa", a=2, z="a")
        self.assertEqual(cls._attrs, {"coucou": dic["coucou"],
                                      "kuku": dic["kuku"]})
        self.assertTrue(cls.__doc__.startswith("haphap\n"))
        self.assertIn("popopos", cls.__doc__)

        # TableBase
        with mock.patch("sqlalchemy.ext.declarative.api"
                        ".DeclarativeMeta.__init__") as dm_patch:
            __init__(cls, "TableBase", (), {})
        self.assertNotEqual(cls.__tablename__, "tablebases")
        dm_patch.assert_not_called()


    def test_query(self):
        """Unit tests for TableMeta.query property."""
        # @property def query(cls)
        query = base.TableMeta.query
        cls = mock.MagicMock(base.TableMeta)
        q = query.fget(cls)
        config.session.query.assert_called_once_with(cls)
        self.assertEqual(q, config.session.query.return_value)
        self.assertIsNone(query.fset)       # read-only
        self.assertIsNone(query.fdel)

    def test_columns(self):
        """Unit tests for TableMeta.columns property."""
        # @property def columns(cls)
        columns = base.TableMeta.columns
        cls = mock.MagicMock(base.TableMeta, __table__=mock.Mock())
        cols = columns.fget(cls)
        self.assertEqual(cols, cls.__table__.columns)
        self.assertIsNone(columns.fset)       # read-only
        self.assertIsNone(columns.fdel)

    def test_attrs(self):
        """Unit tests for TableMeta.attrs property."""
        # @property def attrs(cls)
        attrs = base.TableMeta.attrs
        ats = mock.MagicMock(base.TableMeta, __table__=mock.Mock(),
                             _attrs=mock.Mock())
        cols = attrs.fget(ats)
        self.assertEqual(cols, ats._attrs)
        self.assertIsNone(attrs.fset)       # read-only
        self.assertIsNone(attrs.fdel)

    def test_primary_col(self):
        """Unit tests for TableMeta.primary_col property."""
        # @property def primary_col(cls)
        primary_col = base.TableMeta.primary_col
        cls = mock.MagicMock(base.TableMeta, __table__=mock.Mock())
        # no cols
        cls.__table__.primary_key.columns = {}
        with self.assertRaises(ValueError):
            primary_col.fget(cls)
        # > 1 col
        cls.__table__.primary_key.columns = {"baa", "klklklk"}
        with self.assertRaises(ValueError):
            primary_col.fget(cls)
        # 1 col
        cls.__table__.primary_key.columns = {"bzzzkt"}
        pcol = primary_col.fget(cls)
        self.assertEqual(pcol, "bzzzkt")
        # assert read-only
        self.assertIsNone(primary_col.fset)
        self.assertIsNone(primary_col.fdel)


    @mock.patch("difflib.SequenceMatcher")
    def test_find_nearest(self, sm_patch):
        """Unit tests for TableMeta.find_nearest method."""
        # def find_nearest(cls, chaine, col=None, sensi=0.25, filtre=None,
        #                  solo_si_parfait=True, parfaits_only=True,
        #                  match_first_word=False)
        find_nearest = base.TableMeta.find_nearest
        cls = mock.MagicMock(base.TableMeta, __table__=mock.Mock())
        matcher = sm_patch.return_value

        # bad col name
        cls.columns = {"b": 1, "c": 2}
        with self.assertRaises(ValueError) as cm:
            find_nearest(cls, "qwa", col="a")
        self.assertIn("Colonne 'a' invalide", cm.exception.args[0])
        sm_patch.assert_not_called()
        cls.query.assert_not_called()

        # bad col type
        with self.assertRaises(ValueError) as cm:
            find_nearest(cls, "qwa", col=mock.Mock(sqlalchemy.Column,
                                                   type=sqlalchemy.Integer()))
        self.assertIn("pas de type textuel", cm.exception.args[0])
        sm_patch.assert_not_called()
        cls.query.assert_not_called()

        # no results
        col = mock.Mock(sqlalchemy.Column, type=sqlalchemy.String())
        cls.query.all.return_value = []
        results = find_nearest(cls, "qwa", col=col)
        sm_patch.assert_called_once()
        sm_patch.reset_mock()
        cls.query.filter.assert_not_called()
        cls.query.all.assert_called_once()
        cls.reset_mock()
        self.assertEqual(results, [])

        # no results, filtre
        cls.query.filter.return_value.all.return_value = []
        filtre = mock.Mock()
        results = find_nearest(cls, "qwa", col=col, filtre=filtre)
        sm_patch.assert_called_once()
        sm_patch.reset_mock()
        cls.query.filter.assert_called_once_with(filtre)
        cls.query.all.assert_not_called()
        cls.query.filter.return_value.all.assert_called_once()
        cls.reset_mock()
        self.assertEqual(results, [])

        # one result, exact
        col = mock.Mock(sqlalchemy.Column, key="ab", type=sqlalchemy.String())
        ret = [mock.Mock(ab="Qwà")]
        cls.query.all.return_value = ret
        matcher.ratio.return_value = 1
        results = find_nearest(cls, "Qwà", col=col)
        sm_patch.assert_called_once()
        matcher.set_seq1.assert_called_once_with("qwa")
        matcher.set_seq2.assert_called_once_with("qwa")
        sm_patch.reset_mock()
        cls.query.filter.assert_not_called()
        self.assertEqual(results, [(ret[0], 1)])

        # one result, approx
        matcher.ratio.return_value = 0.337
        results = find_nearest(cls, "qwà", col=col)
        sm_patch.assert_called_once()
        matcher.set_seq1.assert_called_once_with("qwa")
        matcher.set_seq2.assert_called_once_with("qwa")
        sm_patch.reset_mock()
        cls.query.filter.assert_not_called()
        self.assertEqual(results, [(ret[0], 0.337)])

        # one result, under sensi
        matcher.ratio.return_value = 0.337
        results = find_nearest(cls, "qwà", col=col, sensi=0.5)
        sm_patch.assert_called_once()
        matcher.set_seq1.assert_called_once_with("qwa")
        matcher.set_seq2.assert_called_once_with("qwa")
        sm_patch.reset_mock()
        cls.query.filter.assert_not_called()
        self.assertEqual(results, [])

        # no results, col = default
        cls.primary_col = mock.Mock(sqlalchemy.Column, key="gzzt",
                                    type=sqlalchemy.String())
        matcher.ratio.return_value = 0.6
        ret = [mock.Mock(gzzt="Qwà")]
        cls.query.all.return_value = ret
        results = find_nearest(cls, "qwà")
        sm_patch.assert_called_once()
        sm_patch.reset_mock()
        cls.reset_mock()
        self.assertEqual(results, [(ret[0], 0.6)])

        # several results, none exact
        scores = [
            ("qwa", 0.6),
            ("st2", 0.2),
            ("st3", 0.9),
        ]
        ret = [mock.Mock(ab=key) for key, _ in scores]
        cls.query.all.return_value = ret
        matcher.ratio.side_effect = [score for _, score in scores]
        results = find_nearest(cls, "qwà", col=col)
        sm_patch.assert_called_once()
        matcher.set_seq1.assert_called_once_with("qwa")
        self.assertEqual(matcher.set_seq2.call_count, 3)
        matcher.set_seq2.assert_has_calls([mock.call(k) for k, _ in scores])
        sm_patch.reset_mock()
        cls.query.filter.assert_not_called()
        self.assertEqual(results, [(ret[2], 0.9), (ret[0], 0.6)])

        # several results, one exact / solo_si_parfait
        scores = [
            ("qwa", 0.6),
            ("st2", 0.2),
            ("st3", 1),
            ("st4", 0.7),
        ]
        ret = [mock.Mock(ab=key) for key, _ in scores]
        cls.query.all.return_value = ret
        matcher.ratio.side_effect = [score for _, score in scores]
        results = find_nearest(cls, "qwà", col=col)
        sm_patch.assert_called_once()
        matcher.set_seq1.assert_called_once_with("qwa")
        self.assertEqual(matcher.set_seq2.call_count, 3)
        matcher.set_seq2.assert_has_calls([mock.call(k)
                                           for k, _ in scores[:3]])
        sm_patch.reset_mock()
        cls.query.filter.assert_not_called()
        self.assertEqual(results, [(ret[2], 1)])

        # several results, one exact / not solo_si_parfait
        matcher.ratio.side_effect = [score for _, score in scores]
        results = find_nearest(cls, "qwà", col=col, solo_si_parfait=False)
        sm_patch.assert_called_once()
        matcher.set_seq1.assert_called_once_with("qwa")
        self.assertEqual(matcher.set_seq2.call_count, 4)
        matcher.set_seq2.assert_has_calls([mock.call(k) for k, _ in scores])
        sm_patch.reset_mock()
        cls.query.filter.assert_not_called()
        self.assertEqual(results, [(ret[2], 1)])

        # several results, one exact / not solo_si_parfait nor parfaits_only
        matcher.ratio.side_effect = [score for _, score in scores]
        results = find_nearest(cls, "qwà", col=col, solo_si_parfait=False,
                               parfaits_only=False)
        sm_patch.assert_called_once()
        matcher.set_seq1.assert_called_once_with("qwa")
        self.assertEqual(matcher.set_seq2.call_count, 4)
        matcher.set_seq2.assert_has_calls([mock.call(k) for k, _ in scores])
        sm_patch.reset_mock()
        cls.query.filter.assert_not_called()
        self.assertEqual(results, [(ret[2], 1), (ret[3], 0.7), (ret[0], 0.6)])

        # several results, two exacts / solo_si_parfait
        scores = [
            ("qwa", 0.6),
            ("st2", 0.2),
            ("st3", 1),
            ("st4", 0.7),
            ("st5", 1),
            ("st6", 0.1),
        ]
        ret = [mock.Mock(ab=key) for key, _ in scores]
        cls.query.all.return_value = ret
        matcher.ratio.side_effect = [score for _, score in scores]
        results = find_nearest(cls, "qwà", col=col)
        sm_patch.assert_called_once()
        matcher.set_seq1.assert_called_once_with("qwa")
        self.assertEqual(matcher.set_seq2.call_count, 3)
        matcher.set_seq2.assert_has_calls([mock.call(k)
                                          for k, _ in scores[:3]])
        sm_patch.reset_mock()
        cls.query.filter.assert_not_called()
        self.assertEqual(results, [(ret[2], 1)])

        # several results, two exacts / not solo_si_parfait
        matcher.ratio.side_effect = [score for _, score in scores]
        results = find_nearest(cls, "qwà", col=col, solo_si_parfait=False)
        sm_patch.assert_called_once()
        matcher.set_seq1.assert_called_once_with("qwa")
        self.assertEqual(matcher.set_seq2.call_count, 6)
        matcher.set_seq2.assert_has_calls([mock.call(k) for k, _ in scores])
        sm_patch.reset_mock()
        cls.query.filter.assert_not_called()
        self.assertEqual(results, [(ret[2], 1), (ret[4], 1)])

        # several results, two exacts / not solo_si_parfait nor parfaits_only
        matcher.ratio.side_effect = [score for _, score in scores]
        results = find_nearest(cls, "qwà", col=col, solo_si_parfait=False,
                               parfaits_only=False)
        sm_patch.assert_called_once()
        matcher.set_seq1.assert_called_once_with("qwa")
        self.assertEqual(matcher.set_seq2.call_count, 6)
        matcher.set_seq2.assert_has_calls([mock.call(k) for k, _ in scores])
        sm_patch.reset_mock()
        cls.query.filter.assert_not_called()
        self.assertEqual(results, [(ret[2], 1), (ret[4], 1),
                                   (ret[3], 0.7), (ret[0], 0.6)])


        # several results, match_first_word
        scores = [
            ("qwa", 0.6, 0.6),
            ("st2 wiii", 0.2, 1),
            ("st3", 1, 1),
            ("st4 ola", 0.7, 0.9),
            ("st5 ak", 1, 0.7),
            ("st6 hpp", 0.1, 0.6),
            ("st7 hpp", 0.2, 0.1),
            ("st8 hpp", 0.8, 0.1),
        ]
        ret = [mock.Mock(ab=key) for key, *_ in scores]
        cls.query.all.return_value = ret
        matcher.ratio.side_effect = [sc for _, *scs in scores for sc in scs]
        results = find_nearest(cls, "qwà", col=col, match_first_word=True)
        sm_patch.assert_called_once()
        matcher.set_seq1.assert_called_once_with("qwa")
        self.assertEqual(matcher.set_seq2.call_count, 4)
        matcher.set_seq2.assert_has_calls([mock.call(k)
                                           for k, *_ in scores[:2]])
        sm_patch.reset_mock()
        cls.query.filter.assert_not_called()
        self.assertEqual(results, [(ret[1], 1)])

        # several results, match_first_word / two exacts / not solo_si_parfait
        matcher.ratio.side_effect = [sc for _, *scs in scores for sc in scs]
        results = find_nearest(cls, "qwà", col=col, match_first_word=True,
                               solo_si_parfait=False)
        sm_patch.assert_called_once()
        matcher.set_seq1.assert_called_once_with("qwa")
        self.assertEqual(matcher.set_seq2.call_count, 16)
        matcher.set_seq2.assert_has_calls([
            mock.call(sc) for key, *_ in scores
            for sc in [key, key.split(maxsplit=1)[0]]
        ])
        sm_patch.reset_mock()
        cls.query.filter.assert_not_called()
        self.assertEqual(results, [(ret[1], 1), (ret[2], 1), (ret[4], 1)])

        # several results, match_first_word / two exacts /
        #                  not solo_si_parfait nor parfaits_only
        matcher.ratio.side_effect = [sc for _, *scs in scores for sc in scs]
        results = find_nearest(cls, "qwà", col=col, match_first_word=True,
                               solo_si_parfait=False, parfaits_only=False)
        sm_patch.assert_called_once()
        matcher.set_seq1.assert_called_once_with("qwa")
        self.assertEqual(matcher.set_seq2.call_count, 16)
        matcher.set_seq2.assert_has_calls([
            mock.call(sc) for key, *_ in scores
            for sc in [key, key.split(maxsplit=1)[0]]
        ])
        sm_patch.reset_mock()
        cls.query.filter.assert_not_called()
        self.assertEqual(results, [(ret[1], 1), (ret[2], 1), (ret[4], 1),
                                   (ret[3], 0.9), (ret[7], 0.8),
                                   (ret[0], 0.6), (ret[5], 0.6)])



class TestTableBase(unittest.IsolatedAsyncioTestCase):
    """Unit tests for lgrez.bdd.base.TableBase methods."""

    def setUp(self):
        mock_discord.mock_config()

    def tearDown(self):
        mock_discord.unmock_config()

    def test_primary_key(self):
        """Unit tests for TableBase.primary_key method."""
        # @property def primary_key(self)
        primary_key = base.TableBase.primary_key
        class SomeTable(mock.MagicMock):
            primary_col = mock.Mock(key="gzzt")
        slf = SomeTable()   # Table instance
        res = primary_key.fget(slf)
        self.assertEqual(res, slf.gzzt)
        self.assertIsNone(primary_key.fset)       # read-only
        self.assertIsNone(primary_key.fdel)

    def test_update(self):
        """Unit tests for TableBase.update method."""
        # @staticmethod def update()
        update = base.TableBase.update
        update()
        config.session.commit.assert_called_once()

    def test_add(self):
        """Unit tests for TableBase.add method."""
        # def add(self, *other)
        add = base.TableBase.add
        # one
        slf = mock.Mock(base.TableBase)     # Table instance
        add(slf)
        config.session.add.assert_called_once_with(slf)
        slf.update.assert_called_once()
        config.session.reset_mock()
        # several
        slf = mock.Mock(base.TableBase)     # Table instance
        other = tuple(mock.Mock(base.TableBase) for i in range(5))
        add(slf, *other)
        config.session.add.assert_called_once_with(slf)
        config.session.add_all.assert_called_once_with(other)
        slf.update.assert_called_once()
        config.session.reset_mock()

    def test_delete(self):
        """Unit tests for TableBase.delete method."""
        # def delete(self, *other)
        delete = base.TableBase.delete
        # one
        slf = mock.Mock(base.TableBase)     # Table instance
        delete(slf)
        config.session.delete.assert_called_once_with(slf)
        slf.update.assert_called_once()
        config.session.reset_mock()
        # several
        slf = mock.Mock(base.TableBase)     # Table instance
        other = [mock.Mock(base.TableBase) for i in range(5)]
        delete(slf, *other)
        self.assertEqual(config.session.delete.call_count, 6)
        config.session.delete.assert_has_calls(
            [mock.call(item) for item in [slf] + other], any_order=True)
        slf.update.assert_called_once()
        config.session.reset_mock()



"""Loup-Garou de la Rez (lg-rez)

Discord bot for organizing boisterous Werewolf RP games ESPCI-style.

See github.com/loic-simon/lg-rez for informations.
"""

__title__ = "lg-rez"
__author__ = "Loïc Simon, Tom Lacoma"
__license__ = "MIT"
__copyright__ = ("Copyright 2020 - 2021 Loïc Simon & Tom Lacoma - "
                 "Club BD-Jeux × GRIs – ESPCI Paris - PSL")
__version__ = "2.1.3"
__all__ = ["LGBot"]


from lgrez import bot
LGBot = bot.LGBot                # Accès direct à lgrez.LGBot




import asyncio
import functools
from contextlib import ExitStack
from unittest import mock

import discord
from discord.ext import commands
import sqlalchemy

from lgrez import LGBot, config


def _decorate_with_cm(func_or_class, context_manager):
    """Returns the func_or_class wrapped into context_manager."""
    if asyncio.iscoroutinefunction(func_or_class):
        @functools.wraps(func_or_class)
        async def newfunc(*args, **kwargs):
            context_manager.__enter__()
            try:
                await func_or_class(*args, **kwargs)
            finally:
                context_manager.__exit__(None, None, None)
    else:
        @functools.wraps(func_or_class)
        def newfunc(*args, **kwargs):
            context_manager.__enter__()
            try:
                func_or_class(*args, **kwargs)
            finally:
                context_manager.__exit__(None, None, None)

    return newfunc


backup = {
    rc_class.__name__: {attr: rc_class.get_raw(attr) for attr in rc_class}
    for rc_class in [config.Role, config.Channel, config.Emoji]
}

def _prepare_attributes(rc_class, discord_type):
    """Rend prêts les attributs d'une classe ReadyCheck"""
    for attr in rc_class:
        raw = rc_class.get_raw(attr)
        name = raw if isinstance(raw, str) else raw.name
        ready = mock.MagicMock(discord_type, name=name)
        ready.configure_mock(name=name)
        setattr(rc_class, attr, ready)
        backup[rc_class.__name__][attr] = name

def _unprepare_attributes(rc_class):
    """Rend non-prêts les attributs d'une classe ReadyCheck"""
    for attr in rc_class:
        setattr(rc_class, attr, backup[rc_class.__name__][attr])


def mock_config():
    config.bot = mock.MagicMock(LGBot(), command_prefix="!")
    config.session = mock.Mock(sqlalchemy.orm.session.Session)
    config.engine = mock.Mock(sqlalchemy.engine.Engine)
    config.guild = mock.NonCallableMagicMock(discord.Guild)
    config.loop = mock.NonCallableMock()
    _prepare_attributes(config.Role, discord.Role)
    _prepare_attributes(config.Channel, discord.TextChannel)
    _prepare_attributes(config.Emoji, discord.Emoji)

def unmock_config():
    del config.bot, config.session, config.engine, config.guild, config.loop
    _unprepare_attributes(config.Role)
    _unprepare_attributes(config.Channel)
    _unprepare_attributes(config.Emoji)



class TypingMock(mock.AsyncMock):
    def __init__(self, *args, **kwargs):
        super().__init__(discord.context_managers.Typing, *args, **kwargs)

    async def __aenter__(self):
        pass

    async def __aexit__(self, exc_type=None, exc_val=None, exc_tb=None):
        pass


def get_ctx(_command=None, *args, _caller_id=None, **kwargs):
    ctx = mock.NonCallableMagicMock(commands.Context)
    ctx.configure_mock(
        message=mock.NonCallableMagicMock(discord.Message),
        bot=config.bot,
        args=args,
        kwargs=kwargs,
        prefix=config.bot.command_prefix,
        command=_command,
        cog=_command.cog if _command else None,
        guild=config.guild,
        invoked_with=ctx,
        channel=mock.NonCallableMagicMock(discord.TextChannel),
        author=mock.NonCallableMagicMock(discord.Member),
        send=mock.AsyncMock(discord.TextChannel.send),
        typing=mock.MagicMock(discord.TextChannel.typing),
    )
    ctx.typing.return_value = TypingMock()
    ctx.message.configure_mock(
        guild=ctx.guild,
        channel=ctx.channel,
        author=ctx.author,
        webhook_id=None,
    )
    ctx.channel.configure_mock(
        guild=ctx.guild,
        send=ctx.send,
        typing=ctx.typing,
    )
    ctx.author.configure_mock(
        guild=ctx.guild,
        id=_caller_id or -1,
    )
    ctx.new_message = functools.partial(message, ctx)
    ctx.invoke = functools.partial(call, _command, ctx)
    ctx.assert_sent = functools.partial(assert_sent, ctx)
    ctx.assert_not_sent = functools.partial(assert_not_sent, ctx)
    return ctx


async def call(_command, ctx=None, *args, _caller_id=None, cog=None, **kwargs):
    if ctx is None:
        ctx = get_ctx(_command, *args, _caller_id, **kwargs)
    else:
        ctx.args = (*ctx.args, *args)
        ctx.kwargs.update(kwargs)
        if _caller_id is not None:
            ctx.author.configure_mock(id=_caller_id)
    if cog is None:
        cog = _command.cog
    await _command.callback(cog, ctx, *ctx.args, **ctx.kwargs)
    return ctx


def message(ctx, content=None, **kwargs):
    msg = mock.NonCallableMock(discord.Message)
    msg.configure_mock(
        content=content,
        guild=ctx.guild,
        channel=ctx.channel,
        author=ctx.author,
        **kwargs,
    )
    return msg

def chan(name, **kwargs):
    msg = mock.NonCallableMock(discord.TextChannel)
    msg.configure_mock(
        name=name,
        guild=config.guild,
        send=mock.AsyncMock(discord.TextChannel.send),
        typing=mock.MagicMock(discord.TextChannel.typing),
        **kwargs,
    )
    return msg


def assert_sent(chan, *msgs):
    """Raises an AssertionError if chan.send has not been called:

    - len(msgs) times
    - with msgs[<i>] in <i>th message call (string of list of strings, in
        which case each string it contains must be present in the call)
    """
    msgs = list(msgs)
    calls = [call.args[0] for call in chan.send.call_args_list]

    if len(calls) != len(msgs):
        ncalls, nmsgs = len(calls), len(msgs)
        msgs.extend(["[end]"]*(ncalls - nmsgs))
        calls.extend(["[end]"]*(nmsgs - ncalls))

        table = ("     Expected                 | Recieved               \n"
                 "    --------------------------|------------------------\n")
        for msg, call in zip(msgs, calls):
            if msg != "[end]":
                msg = f"*{msg if isinstance(msg, str) else '*'.join(msg)}*"
            called = call.replace("\n", "\\n")
            table += f"     {msg.ljust(24)} | {called} \n"

        raise AssertionError(f"chan.send called {ncalls} times - "
                             f"expecting {nmsgs} calls. Details:\n{table}")

    for msg, call in zip(msgs, calls):
        if isinstance(msg, str):
            msg = [msg]
        for smsg in msg:
            assert str(smsg) in call, (f"chan.send: excepted message '{smsg}' "
                                       f"not found in '{call}'")


def assert_not_sent(chan, *msgs):
    """Raises an AssertionError if chan.send has been called:

    - not len(msgs) times
    - with msgs[<i>] in <i>th message call (string of list of strings, in
        which case each string it contains must NOT be present in the call)
    """
    msgs = list(msgs)
    calls = [call.args[0] for call in chan.send.call_args_list]

    if len(calls) != len(msgs):
        ncalls, nmsgs = len(calls), len(msgs)
        msgs.extend(["[end]"]*(ncalls - nmsgs))
        calls.extend(["[end]"]*(nmsgs - ncalls))

        table = ("     Expected                 | Recieved               \n"
                 "    --------------------------|------------------------\n")
        for msg, call in zip(msgs, calls):
            if msg != "[end]":
                msg = f"*{msg if isinstance(msg, str) else '*'.join(msg)}*"
            called = call.replace("\n", "\\n")
            table += f"     NOT {msg.ljust(20)} | {called} \n"

        raise AssertionError(f"chan.send called {ncalls} times - "
                             f"expecting {nmsgs} calls. Details:\n{table}")

    for msg, call in zip(msgs, calls):
        if isinstance(msg, str):
            msg = [msg]
        for smsg in msg:
            assert str(smsg) not in call, (f"ctx.send: unexcepted message "
                                           f"'{smsg}' found in '{call}'")


def get_member(joueur):
    member = mock.NonCallableMagicMock(discord.Member)
    member.configure_mock(
        guild=config.guild,
        id=joueur.discord_id,
        display_name=joueur.nom,
    )
    return member

def get_chan(joueur):
    chan = mock.NonCallableMagicMock(discord.TextChannel)
    chan.configure_mock(
        guild=config.guild,
        id=joueur.chan_id_,
        name=(config.private_chan_prefix
              + joueur.nom.lower().replace(" ", "-")),
        send=mock.AsyncMock(discord.TextChannel.send),
    )
    return chan


class _MembersAndChansMocker:
    def __init__(self, *joueurs):
        self.ids = {jr.discord_id: get_member(jr) for jr in joueurs}
        self.chans = {jr.chan_id_: get_chan(jr) for jr in joueurs}
        self.members_patch = mock.patch("lgrez.config.guild.get_member",
                                        side_effect=self.ids.get)
        self.chans_patch = mock.patch("lgrez.config.guild.get_channel",
                                      side_effect=self.chans.get)

    def __enter__(self):
        self.members_patch.__enter__()
        self.chans_patch.__enter__()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.members_patch.__exit__(exc_type, exc_val, exc_tb)
        self.chans_patch.__exit__(exc_type, exc_val, exc_tb)

    def __call__(self, func_or_class):
        """Use as a decorator"""
        return _decorate_with_cm(func_or_class, self)


def mock_members_and_chans(*joueurs):
    """Mock corrspondant members and private chans for joueurs.

    Can be used as a context manager or a class/function/async function
    decorator, as for `mock.patch`.
    """
    return _MembersAndChansMocker(*joueurs)


class _Interact:
    funcs = [
        ("lgrez.config.bot", "wait_for"),
        ("lgrez.blocs.tools", "wait_for_message"),
        ("lgrez.blocs.tools", "wait_for_message_here"),
        ("lgrez.blocs.tools", "wait_for_react_clic"),
        # ("lgrez.blocs.tools", "boucle_message"),
        # ("lgrez.blocs.tools", "boucle_query_joueur"),
        ("lgrez.blocs.tools", "yes_no"),
        ("lgrez.blocs.tools", "choice"),
    ]

    def __init__(self, *args):
        self.expected = list(args)
        self.side_effects = iter(args)
        # Itérateur de tuples (func, call) attendus
        self._patchs = {}   # objets _patchs
        self.patchs = {}    # mocks créés à l'enter
        for (mod, func) in self.funcs:
            # Patch de chaque fonction d'interaction
            self._patchs[func] = mock.patch(
                f"{mod}.{func}", new_callable=mock.AsyncMock,
                side_effect=functools.partial(self._side_effect, func),
            )

    def _format_calls(self, msg):
        if len(self.expected) < len(self.calls):
            self.expected.append(("[end]", ""))
        else:
            self.calls.append(("[end]", mock.call()))

        table = ("     Expected                 | Recieved               \n"
                 "    --------------------------|------------------------\n")
        for (efunc, result), (hfunc, call) in zip(self.expected, self.calls):
            expected = f"{efunc}"
            called = f"{hfunc}"   # "{tuple(call.args)}"
            table += f"     {expected.ljust(24)} | {called} \n"
        return f"{msg}\n{table}"

    def _side_effect(self, func, *args, **kwargs):
        self.calls.append((func, mock.call(*args, **kwargs)))
        expected_func, result = next(self.side_effects, (..., ...))
        assert result != ..., self._format_calls(
            "Too many user interactions"
        )
        assert expected_func == func, self._format_calls(
            "Unexpected user interaction"
        )
        if callable(result):        # Function used as side_effect
            return result(*args, **kwargs)
        else:
            return result           # Result value given directly

    def __enter__(self):
        self.stack = ExitStack()
        self.calls = []
        for func in self._patchs:
            mk = self.stack.enter_context(self._patchs[func])
            self.patchs[func] = mk
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.stack.close()
        if exc_type is None:        # no exception
            assert next(self.side_effects, ...) == ..., self._format_calls(
                "Missing user interaction"
            )


    def __call__(self, func_or_class):
        """Use as a decorator"""
        return _decorate_with_cm(func_or_class, self)


def interact(*args):
    return _Interact(*args)



"""lg-rez / bdd / Modèle de données

Déclaration de toutes les tables et leurs colonnes

"""

import discord
import sqlalchemy

from lgrez import config
from lgrez.bdd import base
from lgrez.bdd.base import (autodoc_Column, autodoc_ManyToOne,
                            autodoc_OneToMany, autodoc_DynamicOneToMany,
                            autodoc_ManyToMany)
from lgrez.bdd.enums import ActionTrigger, CibleType


# Tables de jonction (pour many-to-manys)

_baseaction_role = sqlalchemy.Table('_baseactions_roles',
    base.TableBase.metadata,
    sqlalchemy.Column('_role_slug', sqlalchemy.ForeignKey('roles.slug')),
    sqlalchemy.Column('_baseaction_slug',
                      sqlalchemy.ForeignKey('baseactions.slug'))
)


# Tables de données

class Role(base.TableBase):
    """Table de données des rôles.

    Cette table est remplie automatiquement à partir du Google Sheet
    "Rôles et actions" par la commande
    :meth:`\!fillroles <.sync.Sync.Sync.fillroles.callback>`.
    """
    slug = autodoc_Column(sqlalchemy.String(32), primary_key=True,
        doc="Identifiant unique du rôle")

    prefixe = autodoc_Column(sqlalchemy.String(8), nullable=False, default="",
        doc="Article du nom du rôle (``\"Le \"``, ``\"La \"``, ``\"L'\"``...)")
    nom = autodoc_Column(sqlalchemy.String(32), nullable=False,
        doc="Nom (avec casse et accents) du rôle")

    _camp_slug = sqlalchemy.Column(sqlalchemy.ForeignKey("camps.slug"),
        nullable=False)
    camp = autodoc_ManyToOne("Camp", back_populates="roles",
        doc="Camp auquel ce rôle est affilié à l'origine\n\n (On peut avoir "
            "``joueur.camp != joueur.role.camp`` si damnation, passage MV...)")

    actif = autodoc_Column(sqlalchemy.Boolean(), nullable=False, default=True,
        doc="Rôle actif ? (affiché dans la liste des rôles, etc)")

    description_courte = autodoc_Column(sqlalchemy.String(140), nullable=False,
        default="",
        doc="Description du rôle en une ligne")
    description_longue = autodoc_Column(sqlalchemy.String(1800),
        nullable=False, default="",
        doc="Règles et background complets du rôle")

    # to-manys
    joueurs = autodoc_OneToMany("Joueur", back_populates="role",
        doc="Joueurs ayant ce rôle")
    ciblages = autodoc_DynamicOneToMany("Ciblage", back_populates="role",
        doc="Ciblages prenant ce rôle pour cible")
    base_actions = autodoc_ManyToMany("BaseAction", secondary=_baseaction_role,
        back_populates="roles",
        doc="Modèles d'actions associées")

    def __repr__(self):
        """Return repr(self)."""
        return f"<Role '{self.slug}' ({self.prefixe}{self.nom})>"

    def __str__(self):
        """Return str(self)."""
        return self.nom_complet

    @property
    def nom_complet(self):
        """str: Préfixe + nom du rôle"""
        return f"{self.prefixe}{self.nom}"

    @property
    def embed(self):
        """discord.Embed: Embed Discord présentant le rôle et ses actions."""
        emb = discord.Embed(
            title=f"**{self.nom_complet}** – {self.description_courte}",
            description=self.description_longue
        )
        if (emoji := self.camp.discord_emoji_or_none):
            emb.set_thumbnail(url=emoji.url)
        for ba in self.base_actions:
            emb.add_field(name=f"{config.Emoji.action} Action : {ba.slug}",
                          value=ba.temporalite)
        return emb

    @classmethod
    def default(cls):
        """Retourne le rôle par défaut (celui avant attribution).

        Warning:
            Un rôle de :attr:`.slug` :obj:`.config.default_role_slug`
            doit être défini en base.

        Returns:
            ~bdd.Role

        Raises:
            ValueError: rôle introuvable en base
            RuntimeError: session non initialisée
                (:obj:`.config.session` vaut ``None``)
        """
        slug = config.default_role_slug
        role = cls.query.get(slug)
        if not role:
            raise ValueError(
                "Rôle par défaut (de slug "
                f"`lgrez.config.default_role_slug = \"{slug}\"`) non "
                "défini (dans le GSheet Rôles et actions) "
                "ou non chargé (`!fillroles`) !"
            )
        return role


class Camp(base.TableBase):
    """Table de données des camps, publics et secrets.

    Cette table est remplie automatiquement à partir du Google Sheet
    "Rôles et actions" par la commande
    :meth:`\!fillroles <.sync.Sync.Sync.fillroles.callback>`.
    """
    slug = autodoc_Column(sqlalchemy.String(32), primary_key=True,
        doc="Identifiant unique du camp")

    nom = autodoc_Column(sqlalchemy.String(32), nullable=False,
        doc="Nom (affiché) du camp")
    description = autodoc_Column(sqlalchemy.String(1000), nullable=False,
        default="",
        doc="Description (courte) du camp")

    public = autodoc_Column(sqlalchemy.Boolean(), nullable=False, default=True,
        doc="L'existance du camp (et des rôles liés) est connue de tous ?")

    emoji = autodoc_Column(sqlalchemy.String(32),
        doc="Nom de l'emoji associé au camp (doit être le nom d'un "
            "emoji existant sur le serveur)")

    # One-to-manys
    joueurs = autodoc_OneToMany("Joueur", back_populates="camp",
        doc="Joueurs appartenant à ce camp")
    roles = autodoc_OneToMany("Role", back_populates="camp",
        doc="Rôles affiliés à ce camp de base")
    ciblages = autodoc_DynamicOneToMany("Ciblage", back_populates="camp",
        doc="Ciblages prenant ce camp pour cible")

    def __repr__(self):
        """Return repr(self)."""
        return f"<Camp '{self.slug}' ({self.nom})>"

    def __str__(self):
        """Return str(self)."""
        return str(self.nom)

    @property
    def discord_emoji(self):
        """discord.Emoji: Emoji Discord correspondant à ce camp

        Raises:
            ValueError: :attr:`.emoji` non défini ou manquant sur le serveur
            ~ready_check.NotReadyError: bot non connecté
                (:obj:`.config.guild` vaut ``None``)
        """
        if not self.emoji:
            raise ValueError(f"{self}.emoji non défini !")

        try:
            return next(e for e in config.guild.emojis if e.name == self.emoji)
        except StopIteration:
            raise ValueError(f"Pas d'emoji :{self.emoji}: "
                             "sur le serveur !") from None

    @property
    def discord_emoji_or_none(self):
        """:class:`discord.Emoji` | ``None``: :attr:`.discord_emoji` si défini

        Raises:
            ~ready_check.NotReadyError: bot non connecté
                (:obj:`.config.guild` vaut ``None``)
        """
        try:
            return self.discord_emoji
        except ValueError:
            return None

    @classmethod
    def default(cls):
        """Retourne le camp par défaut (celui avant attribution).

        Warning:
            Un camp de :attr:`.slug` :obj:`.config.default_camp_slug`
            doit être défini en base.

        Returns:
            ~bdd.Camp

        Raises:
            ValueError: camp introuvable en base
            RuntimeError: session non initialisée
                (:obj:`.config.session` vaut ``None``)
        """
        slug = config.default_camp_slug
        camp = cls.query.get(slug)
        if not camp:
            raise ValueError(
                "Camp par défaut (de slug "
                f"lgrez.config.default_camp_slug = \"{slug}\") non "
                "défini (dans le GSheet Rôles et actions) ou non "
                f"chargé (`!fillroles`) !"
            )
        return camp


class BaseAction(base.TableBase):
    """Table de données des actions définies de base (non liées à un joueur).

    Cette table est remplie automatiquement à partir du Google Sheet
    "Rôles et actions" par la commande :meth:`\!fillroles
    <.remplissage_bdd.RemplissageBDD.RemplissageBDD.fillroles.callback>`.
    """
    slug = autodoc_Column(sqlalchemy.String(32), primary_key=True,
        doc="Identifiant unique de l'action")

    trigger_debut = autodoc_Column(sqlalchemy.Enum(ActionTrigger),
        nullable=False, default=ActionTrigger.perma,
        doc="Mode de déclenchement de l'ouverture de l'action")
    trigger_fin = autodoc_Column(sqlalchemy.Enum(ActionTrigger),
        nullable=False, default=ActionTrigger.perma,
        doc="Mode de déclenchement de la clôture de l'action")
    instant = autodoc_Column(sqlalchemy.Boolean(), nullable=False,
        default=False,
        doc="L'action est instantannée (conséquence dès la prise de décision)"
            " ou non (conséquence à la fin du créneau d'action)")

    heure_debut = autodoc_Column(sqlalchemy.Time(),
        doc="Si :attr:`.trigger_debut` vaut "
            ":attr:`~ActionTrigger.temporel`, l'horaire associé")
    heure_fin = autodoc_Column(sqlalchemy.Time(),
        doc="Si :attr:`.trigger_fin` vaut\n"
            "- :attr:`~ActionTrigger.temporel` : l'horaire associé ;\n"
            "- :attr:`~ActionTrigger.delta`, l'intervalle associé")

    base_cooldown = autodoc_Column(sqlalchemy.Integer(), nullable=False,
        default=0,
        doc="Temps de rechargement entre deux utilisations du pouvoir "
            "(``0`` si pas de cooldown)")
    base_charges = autodoc_Column(sqlalchemy.Integer(),
        doc="Nombre de charges initiales du pouvoir (``None`` si illimité)")
    refill = autodoc_Column(sqlalchemy.String(32), nullable=False, default="",
        doc="Évènements pouvant recharger l'action, séparés par des virgules "
            "(``\"weekends\"``, ``\"forgeron\"``, ``\"rebouteux\"``...)")

    lieu = autodoc_Column(sqlalchemy.String(100),
        doc="*Attribut informatif, non exploité dans la version actuelle "
            "(Distance/Physique/Lieu/Contact/Conditionnel/None/Public)*")
    interaction_notaire = autodoc_Column(sqlalchemy.String(100),
        doc="*Attribut informatif, non exploité dans la version actuelle "
            "(Oui/Non/Conditionnel/Potion/Rapport ; None si récursif)*")
    interaction_gardien = autodoc_Column(sqlalchemy.String(100),
        doc="*Attribut informatif, non exploité dans la version actuelle "
            "(Oui/Non/Conditionnel/Taverne/Feu/MaisonClose/Précis/"
            "Cimetière/Loups ; None si récursif)*")
    mage = autodoc_Column(sqlalchemy.String(100),
        doc="*Attribut informatif, non exploité dans la version actuelle "
            "(Oui/Non/Changement de cible/...)*")

    decision_format = autodoc_Column(sqlalchemy.String(200),
        nullable=False, default="",
        doc="Description des utilisations de ces action, sous forme de "
            "texte formaté avec les noms des :attr:`.BaseCiblage.slug` "
            "entre accolades (exemple : ``Tuer {cible}``)")

    # -to-manys
    actions = autodoc_OneToMany("Action", back_populates="base",
        doc="Actions déroulant de cette base")
    base_ciblages = autodoc_OneToMany("BaseCiblage",
        back_populates="base_action", cascade="all, delete-orphan",
        order_by="BaseCiblage.prio",
        doc="Ciblages de ce modèle d'action (triés par priorité)")
    roles = autodoc_ManyToMany("Role", secondary=_baseaction_role,
        back_populates="base_actions",
        doc="Rôles ayant cette action de base")

    def __repr__(self):
        """Return repr(self)."""
        return f"<BaseAction '{self.slug}'>"

    def __str__(self):
        """Return str(self)."""
        return str(self.slug)

    @property
    def temporalite(self):
        """str: Phrase décrivant le mode d'utilisation / timing de l'action."""
        def _time_to_heure(tps):
            if not tps:
                return ""
            if tps.hour == 0:
                return f"{tps.minute} min"
            if tps.minute > 0:
                return f"{tps.hour}h{tps.minute:02}"
            return f"{tps.hour}h"

        rep = ""
        # Périodicté
        if self.trigger_debut == ActionTrigger.perma:
            rep += "N'importe quand"
        elif self.trigger_debut == ActionTrigger.start:
            rep += "Au lancement de la partie"
        elif self.trigger_debut == ActionTrigger.mort:
            rep += "À la mort"
        else:
            if self.base_cooldown:
                rep += f"Tous les {self.base_cooldown + 1} jours "
            else:
                rep += "Tous les jours "

            # Fenêtre
            if self.trigger_debut == ActionTrigger.mot_mjs:
                rep += "à l'annonce des résultats du vote"
            elif self.trigger_debut == ActionTrigger.open_cond:
                rep += "pendant le vote condamné"
            elif self.trigger_debut == ActionTrigger.open_maire:
                rep += "pendant le vote pour le maire"
            elif self.trigger_debut == ActionTrigger.open_loups:
                rep += "pendant le vote des loups"
            elif self.trigger_debut == ActionTrigger.close_cond:
                rep += "à la fermeture du vote condamné"
            elif self.trigger_debut == ActionTrigger.close_maire:
                rep += "à la fermeture du vote pour le maire"
            elif self.trigger_debut == ActionTrigger.close_loups:
                rep += "à la fermeture du vote des loups"
            elif self.trigger_debut == ActionTrigger.temporel:
                if self.trigger_fin == ActionTrigger.temporel:
                    rep += f"de {_time_to_heure(self.heure_debut)}"
                else:
                    rep += f"à {_time_to_heure(self.heure_debut)}"

        # Fermeture
        if self.trigger_fin == ActionTrigger.delta:
            rep += f" – {_time_to_heure(self.heure_fin)} pour agir"
        elif self.trigger_fin == ActionTrigger.temporel:
            rep += f" à {_time_to_heure(self.heure_fin)}"

        # Autres caractères
        if self.instant:
            rep += f" (conséquence instantanée)"
        if self.base_charges:
            rep += f" – {self.base_charges} fois"
        if "weekends" in self.refill:
            rep += f" par semaine"

        return rep


class BaseCiblage(base.TableBase):
    """Table de données des modèles de ciblages des actions de base.

    [TODO] Cette table est remplie automatiquement à partir du Google Sheet
    "Rôles et actions" par la commande :meth:`\!fillroles
    <.remplissage_bdd.RemplissageBDD.RemplissageBDD.fillroles.callback>`.
    """
    _id = autodoc_Column(sqlalchemy.Integer(), primary_key=True,
        doc="Identifiant unique du modèle de ciblage, sans signification")

    _baseaction_slug = sqlalchemy.Column(sqlalchemy.ForeignKey(
        "baseactions.slug"), nullable=False)
    base_action = autodoc_ManyToOne("BaseAction",
        back_populates="base_ciblages",
        doc="Modèle d'action définissant ce ciblage")

    slug = autodoc_Column(sqlalchemy.String(32), nullable=False,
        default="unique",
        doc="Identifiant de la cible dans le modèle d'action")
    type = autodoc_Column(sqlalchemy.Enum(CibleType), nullable=False,
        default=CibleType.texte,
        doc="Message d'interaction au joueur au moment de choisir la cible")

    prio = autodoc_Column(sqlalchemy.Integer(), nullable=False, default=1,
        doc="Ordre (relatif) d'apparition du ciblage lors du ``!action`` "
        "\n\nSi deux ciblages ont la même priorité, ils seront considérés "
        "comme ayant une signification symmétrique (notamment, si "
        ":attr:`doit_changer` vaut ``True``, tous les membres du groupe "
        "devront changer) ; l'ordre d'apparition dépend alors de leur "
        ":attr:`slug`, par ordre alphabétique (``cible1`` < ``cible2``).")

    phrase = autodoc_Column(sqlalchemy.String(1000), nullable=False,
        default="Cible ?",
        doc="Message d'interaction au joueur au moment de choisir la cible")

    obligatoire = autodoc_Column(sqlalchemy.Boolean(), nullable=False,
        default=True,
        doc="Si le ciblage doit obligatoirement être renseigné")
    doit_changer = autodoc_Column(sqlalchemy.Boolean(), nullable=False,
        default=False,
        doc="Si la cible doit changer d'une utilisation à l'autre.\n\n"
        "Si la dernière utilisation est ignorée ou contrée, il n'y a "
        "pas de contrainte.")

    # one-to-manys
    ciblages = autodoc_DynamicOneToMany("Ciblage", back_populates="base",
        doc="Ciblages déroulant de cette base")

    def __repr__(self):
        """Return repr(self)."""
        return f"<BaseCiblage #{self._id} ({self.base_action}/{self.slug})>"



"""lg-rez / Variables globales

Personalisation de différents paramètres et accès global

"""

import discord

from lgrez.blocs import ready_check


#: str: Préfixe des noms des salons de conversation bot.
private_chan_prefix = "conv-bot-"

#: str: Nom de la catégorie des conversations bot, pour l'inscription
#: (sera éventuellement suivi de 2, 3... si plus de 50 joueurs).
private_chan_category_name = "CONVERSATION BOT"

#: str: Nom de la catégorie des boudoirs
#: (sera éventuellement suivi de 2, 3... si plus de 50 boudoirs).
boudoirs_category_name = "BOUDOIRS"


#: str: Date de début de saison (pour information lors de l'inscription).
debut_saison = "32 plopembre"

#: bool: Si ``False``, le processus d'insciption ne demandera pas la
#: chambre des joueurs, qui seront tous inscrits en :attr:`chambre_mj`
#: (et la chambre ne sera pas indiquée dans ``!vivants``).
demande_chambre = True

#: str: Nom par défaut de la :attr:`~.bdd.Joueur.chambre` des joueurs.
chambre_mj = "[chambre MJ]"

async def additional_inscription_step(member, chan):
    """Coroutine permettant d'ajouter des étapes au processus d'inscription.

    Cette coroutine est appelée par :func:`.features.inscription.main`
    juste avant l'inscription en base. Si elle renvoie `False`,
    l'inscription est annulée ; si elle ne renvoie rien ou une autre
    valeur, elle continue selon le processus habituel.

    Args:
        member (discord.Member): membre en cours d'inscription.
        chan (discord.TextChannel): chan perso créé pour l'inscription.

    Returns:
        Si ``False``, annule l'inscription.
    """
    pass


#: bool: Si ``True``, le bot appellera :meth:`.LGBot.i_am_alive` toutes
#: les 60 secondes. Ce n'est pas activé par défaut.
output_liveness = False


#: str: :attr:`~.bdd.Role.slug` du rôle par défaut, attribué aux
#: joueurs lors de l'inscription (renvoyé par :meth:`.bdd.Role.default`).
#: Doit correspodre à un rôle existant (défini dans le GSheet *Rôles et
#: actions*).
default_role_slug = "nonattr"

#: str: :attr:`~.bdd.Camp.slug` du camp par défaut, attribué aux
#: joueurs lors de l'inscription (renvoyé par :meth:`.bdd.Camp.default`).
#: Doit correspodre à un camp existant (défini dans le GSheet *Rôles et
#: actions*).
default_camp_slug = "nonattr"


#: str: Nom de la feuille du *Tableau de bord* contenant l'état actuel
#: des joueurs, sur laquelle sont effectuées les modifications.
tdb_main_sheet = "Journée en cours"

#: str: Nom de la feuille du *Tableau de bord* contenant les résultats
#: des votes (après corrections manuelles éventuelles).
tdb_votes_sheet = "Journée en cours"

#: int: Numéro de la ligne de la feuille principale
#: (:attr:`~lgrez.config.tdb_main_sheet`)
#: du *Tableau de bord* contenant les noms des colonnes (commençant de 1).
tdb_header_row = 3

#: str: Nom de la colonne de la feuille principale
#: (:attr:`~lgrez.config.tdb_main_sheet`)
#: du *Tableau de bord* contenant les IDs Discord des joueurs.
tdb_id_column = "A"

#: tuple[str]: Noms de la première et de la dernière colonne de la zone de
#: la feuille principale (:attr:`~lgrez.config.tdb_main_sheet`) du *Tableau
#: de bord* contenant les informations (colonnes de la BDD) des joueurs.
tdb_main_columns = ("J", "Q")

#: tuple[str]: Noms de la première et de la dernière colonne de la zone de
#: la feuille principale (:attr:`~lgrez.config.tdb_main_sheet`) du *Tableau
#: de bord* contenant l'ancien état des informations des joueurs
#: (avant ``!sync``).
tdb_tampon_columns = ("B", "I")


#: int: Nombre maximal de modèles de ciblages (:class:`.bdd.BaseCiblage`)
#: renseignés pour chaque modèle d'action (:class:`.bdd.BaseAction`), à
#: droite de la feuille ``baseactions`` du GSheet *Rôles et actions*.
max_ciblages_per_action = 3


#: str: :attr:`.bdd.BaseAction.slug` de l'action de base permettant
#: de modifier un vote (rôle de l'*Intigant* dans le jeu PCéen).
#: Cette baseaction doit avoir deux ciblages de slugs "cible" et "vote".
modif_vote_baseaction = "modification-vote"

#: str: :attr:`.bdd.BaseAction.slug` de l'action de base permettant
#: d'ajouter un/des vote(s) (rôle du *Corbeau* dans le jeu PCéen).
ajout_vote_baseaction = "ajout-vote"

#: int: Nombre de votes ajoutés par l'action :attr:`ajout_vote_baseaction`.
n_ajouts_votes = 2


#: str: Nom de la feuille du GSheet *Données brûtes* où enregistrer
#: les votes brutes pour le condamné du jour.
db_votecond_sheet = "votecond_brut"

#: str: Nom de la feuille du GSheet *Données brûtes* où enregistrer
#: les votes brutes pour le nouveau maire.
db_votemaire_sheet = "votemaire_brut"

#: str: Nom de la feuille du GSheet *Données brûtes* où enregistrer
#: les votes brutes pour le vote des loups.
db_voteloups_sheet = "voteloups_brut"

#: str: Nom de la feuille du GSheet *Données brûtes* où enregistrer
#: les actions effectuées.
db_actions_sheet = "actions_brut"


#: list[str]: Mots-clés (en minuscule) utilisables (quelque soit la casse)
#: pour arrêter une commande en cours d'exécution.
stop_keywords = ["stop", "!stop"]


#: list[str]: Mots-clés de rechargement (dans :attr:`.bdd.BaseAction.refill`)
#: permettant de recharger une action à son nombre de charges initial.
refills_full = ["weekends"]

#: list[str]: Mots-clés de rechargement (dans :attr:`.bdd.BaseAction.refill`)
#: permettant de recharger une action de une charge.
refills_one = ["forgeron", "rebouteux", "divin"]

#: list[str]: Mots-clés de rechargement (dans :attr:`.bdd.BaseAction.refill`)
#: à utiliser par le MJ pour ajouter une charge à une action.
refills_divins = ["divin"]



class Role(ready_check.ReadyCheck, check_type=discord.Role):
    """Rôles Discord nécessaires au jeu

    Cette classe dérive de :class:`.ready_check.ReadyCheck` :
    accéder aux attributs ci-dessous avant que le bot ne soit connecté
    au serveur lève une :exc:`~.ready_check.NotReadyError`.

    Plus précisément, :meth:`.LGBot.on_ready` remplace le nom du rôle
    par l'objet :class:`discord.Role` correspondant : si les noms des
    rôles sur Discord ont été modifiés, indiquer leur nom ici
    (``lgrez.config.Role.x = "nouveau nom"``) avant de lancer le bot,
    sans quoi :meth:`.LGBot.on_ready` lèvera une erreur.

    **Ne pas instancier cette classe.**

    Rôles utilisés (dans l'ordre hiérarchique conseillé) :

    Attributes:
        mj: Maître du Jeu.
            Nom par défaut : "MJ".
        joueur_en_vie: Joueur vivant, pouvant parler publiquement.
            Nom par défaut : "Joueur en vie".
        joueur_mort: Joueur mort, ne pouvant pas parler publiquement.
            Nom par défaut : "Joueur mort".
        maire: Joueur élu Maire, mis en avant et pouvant utiliser @everyone.
            Nom par défaut : "Maire".
        redacteur: Rôle permettant à un joueur d'utiliser les commandes de
            gestion d'IA (voir :class:`features.gestion_ia.GestionIA`). Mettre
            le même nom que le rôle des MJs si vous voulez supprimer ce rôle.
            Nom par défaut : "Rédacteur".
        everyone: Rôle de base. Les joueurs dont le rôle le plus élevé
            est ce rôle (ou moins) seront ignorés par le bot.
            Nom par défaut: "@everyone" (rôle Discord de base)
    """
    mj = "MJ"
    redacteur = "Rédacteur"
    joueur_en_vie = "Joueur en vie"
    joueur_mort = "Joueur mort"
    maire = "Maire"
    everyone = "@everyone"


class Channel(ready_check.ReadyCheck, check_type=discord.TextChannel):
    """Salons Discord nécessaires au jeu

    Cette classe dérive de :class:`.ready_check.ReadyCheck` : accéder
    aux attributs ci-dessous avant que le bot ne soit connecté au
    serveur lève une :exc:`~.ready_check.NotReadyError`.

    Plus précisément, :meth:`.LGBot.on_ready` remplace le nom du rôle
    par l'objet :class:`discord.TextChannel` correspondant : si les noms
    des salons sur Discord ont été modifiés, indiquer leur nom ici
    (``lgrez.config.Channel.x = "nouveau nom"``) avant de lancer le bot,
    sans quoi :meth:`.LGBot.on_ready` lèvera une erreur.

    **Ne pas instancier cette classe.**

    Salons utilisés (dans l'ordre d'affichage conseillé) :

    Attributes:
        roles: Salon listant les rôles (catégorie Informations).
            Nom par défaut : "roles".
        logs: Salon pour les messages techniques (catégorie réservée aux MJs).
            Nom par défaut : "logs".
        annonces: Salon d'annonces (catégorie Place du village).
            Nom par défaut : "annonces".
        haros: Salon des haros et candidatures (catégorie Place du village).
            Nom par défaut : "haros".
        debats: Salon de discussion principal (catégorie Place du village).
            Nom par défaut : "débats".
    """
    roles = "rôles"
    logs = "logs"
    annonces = "annonces"
    haros = "haros"
    debats = "débats"


class Emoji(ready_check.ReadyCheck, check_type=discord.Emoji):
    """Emojis Discord nécessaires au jeu

    Cette classe dérive de :class:`.ready_check.ReadyCheck` : accéder
    aux attributs ci-dessous avant que le bot ne soit connecté au
    serveur lève une :exc:`~.ready_check.NotReadyError`.

    Plus précisément, :meth:`.LGBot.on_ready` remplace le nom du rôle
    par l'objet :class:`discord.Emoji` correspondant : si les noms
    des emojis sur Discord ont été modifiés, indiquer leur nom ici
    (``lgrez.config.Emoji.x = "nouveau nom"``) avant de lancer le bot,
    sans quoi :meth:`.LGBot.on_ready` lèvera une erreur.

    **Ne pas instancier cette classe.**

    Emojis utilisés (noms par défaut identiques aux noms des attributs) :

    Attributes:
        ha
        ro: Accolés, forment le mot « haro »
        bucher: Représente le vote pour le condamné du jour
        maire: Représente le vote pour le nouveau maire
        lune: Représente le vote des loups
        action: Représente les actions de rôle
        void: Image vide, pour séparations verticales et autres filouteries
    """
    ha = "ha"
    ro = "ro"
    bucher = "bucher"
    maire = "maire"
    action = "action"
    lune = "lune"
    void = "void"


class _ModuleGlobals(ready_check.ReadyCheck):
    """Module-level attributes with not-None ReadyCheck

    (attributes accessed by __getattr__, documented directly in config.rst)
    """
    guild = None
    bot = None
    loop = None
    engine = None
    session = None

    webhook = None


# Variable interne, pour suivi des objets manquants (ne pas changer)
_missing_objects = 0


# Called when module attribute not found: try to look in _ModuleGlobals
def __getattr__(attr):
    try:
        return getattr(_ModuleGlobals, attr)
    except AttributeError:
        raise AttributeError(
            f"module '{__name__}' has no attribute '{attr}'"
        ) from None
